from dreem.vector.main import run
from dreem.util.cli import *


@click.command()
@opt_top_dir
@opt_library
@opt_coords
@opt_primers
@opt_fill
@opt_parallel
@opt_min_phred
@opt_phred_enc
@opt_rerun
@arg_fasta
@arg_bams
@opt_cpus
def cli(*args, **opts):
    run(*args, **opts)


if __name__ == '__main__':
    cli()



import numpy as np

def compute_wilson_interval(p, n, z = 1.96):
    denominator = 1 + z**2/n
    centre_adjusted_probability = p + z*z / (2*n)
    adjusted_standard_deviation = np.sqrt((p*(1 - p) + z*z / (4*n)) / n)
    
    lower_bound = (centre_adjusted_probability - z*adjusted_standard_deviation) / denominator
    upper_bound = (centre_adjusted_probability + z*adjusted_standard_deviation) / denominator
    return (lower_bound, upper_bound)




from __future__ import annotations

import logging
from collections import defaultdict, namedtuple
from functools import cached_property
from itertools import repeat, starmap
import os
import pathlib
import re
from datetime import datetime
from hashlib import file_digest
from multiprocessing import Pool
import time
from typing import Any, ClassVar

import numpy as np
import pandas as pd
from pydantic import (BaseModel, Extra, Field, NonNegativeInt, NonNegativeFloat, PositiveInt,
                      StrictStr, validator, root_validator)

from dreem.util.cli import ParallelChoice
from dreem.util import path
from dreem.util.seq import DNA, FastaParser
from dreem.vector.samview import SamViewer
from dreem.vector.vector import SamRecord


DEFAULT_BATCH_SIZE = 33_554_432  # 2^25 bytes ≈ 33.6 Mb
DEFAULT_MIN_PHRED = 25  # minimum Phred score to consider a base in a FASTQ file


RegionTuple = namedtuple("PrimerTuple", ["pos5", "pos3"])


class Region(object):
    """
    Represent a region of a reference sequence between two coordinates.

    Attributes
    ----------
    ref_seq: DNA
        Entire reference sequence (not of just the region).
    ref: str
        Name of the reference sequence.
    end5: int (1 ≤ end5 ≤ end3)
        Coordinate of the reference sequence at which the region's 5' end is
        located (1-indexed).
    end3: int (end5 ≤ end3 ≤ len(ref_seq))
        Coordinate of the reference sequence at which the region's 3' end is
        located (1-indexed; end3 itself is included in the region).

    Examples
    --------
    >>> seq = DNA(b"CATCTGGA")
    >>> name = "example"
    >>> region = Region(ref_seq=seq, ref=name, end5=1, end3=8)
    >>> assert region.region_seq == seq
    >>> region = Region(ref_seq=seq, ref=name, end5=1, end3=-1)
    >>> assert region.region_seq == seq
    >>> region = Region(ref_seq=seq, ref=name, end5=-8, end3=8)
    >>> assert region.region_seq == seq
    >>> region = Region(ref_seq=seq, ref=name, end5=3, end3=7)
    >>> assert region.region_seq == DNA(b"TCTGG")
    >>> region = Region(ref_seq=seq, ref=name, end5=-5, end3=-3)
    >>> assert region.region_seq == DNA(b"CTG")
    >>> try:
    ...     region = Region(ref_seq=seq, ref=name, end5=-9, end3=5)
    ...     assert False, "Failed to catch end5 < -len(ref_seq)"
    ... except ValueError:
    ...     pass
    >>> try:
    ...     region = Region(ref_seq=seq, ref=name, end5=6, end3=5)
    ...     assert False, "Failed to catch end3 < end5"
    ... except ValueError:
    ...     pass
    >>> try:
    ...     region = Region(ref_seq=seq, ref=name, end5=1, end3=9)
    ...     assert False, "Failed to catch end3 > len(ref_seq)"
    ... except ValueError:
    ...     pass
    """

    def __init__(self, *, ref_seq: DNA, ref: str, end5: int, end3: int):
        """
        Parameters
        ----------
        ref_seq: DNA
            Entire reference sequence (not of just the region).
        ref: str
            Name of the reference sequence.
        end5: int (-len(ref_seq) ≤ end5 ≤ len(ref_seq); end5 ≠ 0)
            Coordinate of the reference sequence at which the 5' end of the
            region is located. If positive, number the coordinates of the
            reference sequence 1, 2, ... starting at the 5' end (i.e. 1-based
            indexing). If negative, number the coordinates of the reference
            sequence -1, -2, ... starting at the 3' end (i.e. 1-based indexing
            from the other side), then convert to the corresponding (positive)
            1-based index from the 5' end.
        end3: int (-len(ref_seq) ≤ end3 ≤ len(ref_seq); end3 ≠ 0)
            Coordinate of the reference sequence at which the region's 3' end is
            located. Follows the same coordinate numbering convention as end5.
        """
        length = len(ref_seq)
        if end5 < 0:
            # If end5 is negative, find the corresponding positive coordinate.
            end5 += length + 1
        if end3 < 0:
            # If end3 is negative, find the corresponding positive coordinate.
            end3 += length + 1
        if not 1 <= end5 <= end3 <= length:
            raise ValueError("Must have 1 ≤ end5 ≤ end3 ≤ len(ref_seq), "
                             f"but got end5 = {end5}, end3 = {end3}, and "
                             f"len(ref_seq) = {length}")
        self.ref_seq = ref_seq
        self.ref = ref
        self.end5 = end5
        self.end3 = end3

    @property
    def path_fields(self):
        return {"ref": self.ref, "end5": self.end5, "end3": self.end3}

    @property
    def prof_fields(self):
        return {"ref": self.ref, "end5": self.end5, "end3": self.end3}
    
    @property
    def spanning(self) -> bool:
        """ Return whether the region spans the entire reference sequence. """
        return self.end5 == 1 and self.end3 == len(self.ref_seq)
    
    @property
    def region_seq(self) -> DNA:
        """ Return the sequence of the region of interest. """
        return self.ref_seq[self.end5 - 1: self.end3]

    @property
    def length(self):
        """ Return the length of the region of interest. """
        return self.end3 - self.end5 + 1

    @property
    def positions(self):
        """ Return all positions in the region of interest as an NDArray. """
        return np.arange(self.end5, self.end3 + 1)

    @property
    def ref_coords(self):
        """ Return the name of the reference and the first and last positions
        of the region of interest; for equality testing and hashing. """
        return self.ref, self.end5, self.end3
    
    @property
    def columns(self):
        """ Return a tuple of the bases and coordinates in the region, each of
        the form '{base}{position}' (e.g. ['G13', 'C14', 'A15']). """
        return list(f"{chr(base)}{coord}" for base, coord
                    in zip(self.region_seq, self.positions, strict=True))


class RegionFinder(Region):
    """
    The 5' and 3' ends of a region can be given explicitly as integers, but if
    the sample is of an amplicon (i.e. generated by RT-PCR using site-specific
    primers), then it is often more convenient to enter the sequences of the
    PCR primers and have the software determine the coordinates. RegionFinder
    accepts 5' and 3' coordinates given as integers or primers, validates them,
    and stores the coordinates as integers. This process works as follows:

    end5
    - If given as a parameter, this value is used.
    - Else if fwd is given, first is based on the location of its 3' end.
      - Else first is set to 1.
        - last
          - If last is given as an argument, last is set to this value.
          - Else if rev is given, last is based on the location of the 5' end
            of its reverse complement.
          - Else last is set to the length of ref_seq.

    """

    def __init__(self, *, ref_seq: DNA | None, ref: str,
                 end5: int | None = None, end3: int | None = None,
                 fwd: DNA | None = None, rev: DNA | None = None,
                 primer_gap: int = 2):
        """
        Parameters
        ----------
        ref_seq: see superclass
        ref: see superclass
        end5: see superclass
        end3: see superclass
        fwd: DNA | None = None (optional)
            (For amplicons only) Sequence of the forward PCR primer that was
            used to generate the amplicon
        rev: DNA | None = None (optional)
            (For amplicons only) Sequence of the reverse PCR primer that was
            used to generate the amplicon
        primer_gap: int = 2 (optional)
            (For coordinates specified by fwd/rev only) Number of positions 3'
            of the forward primer and 5' of the reverse primer to exclude from
            the region. Coordinates within 1 - 2 nucleotides of each primer may
            contain DMS reactivity artifacts. If primer_gap = 0, then end5 and
            end3 are set, respectively, to the coordinates immediately adjacent
            to (i.e. 1 nucleotide 3' and 5' of) the 3' end of the forward and
            reverse primers. The default (primer_gap = 2) generally works well.
        """
        if ref_seq is None:
            raise ValueError(f"No sequence for reference named '{ref}'")
        offset = primer_gap + 1
        if end5 is None:
            # If first is to be determined from the fwd primer sequence,
            # the primer is aligned to the reference, and first is set to
            # the position (primer_gap + 1) downstream of its 3' coordinate.
            end5 = (1 if fwd is None
                    else self.locate(ref_seq, fwd).pos3 + offset)
        if end3 is None:
            # If last is to be determined from the rev primer sequence,
            # the reverse complement of the primer is aligned to the reference,
            # and last is set to the position (primer_gap + 1) upstream of its
            # 5' coordinate.
            end3 = (len(ref_seq) if rev is None
                    else self.locate(ref_seq, rev.rc).pos5 - offset)
        super().__init__(ref_seq=ref_seq, ref=ref, end5=end5, end3=end3)

    @staticmethod
    def locate(ref_seq: DNA, primer: DNA) -> RegionTuple:
        """
        Return the 5' and 3' positions (1-indexed) of a primer within a
        reference sequence. The primer must occur exactly once in the
        reference, otherwise an error is raised.

        Parameters
        ----------
        ref_seq: DNA
            Sequence of the entire reference (not just the region of interest)
        primer: DNA
            Sequence of the forward PCR primer or of the reverse complement of
            the reverse PCR primer
        
        Returns
        -------
        RegionTuple
            Named tuple of the first and last positions that the primer occupies
            in the reference sequence. Positions are 1-indexed and include the
            first and last coordinates.
        """
        matches = list(re.finditer(primer, ref_seq))
        if not matches:
            raise ValueError(f"Primer '{primer}' is not in ref '{ref_seq}'")
        if len(matches) > 1:
            raise ValueError(f"Primer '{primer}' occurs {len(matches)} times "
                             f"in ref '{ref_seq}'")
        # Add 1 to convert from 0-indexed (re.finditer) to 1-indexed (DREEM).
        pos5 = matches[0].start() + 1
        # No change is needed to convert from exclusive 0-indexed (re.finditer)
        # to inclusive 1-indexed (DREEM).
        pos3 = matches[0].end()
        return RegionTuple(pos5, pos3)
    

class MutationalProfile(Region):
    """
    Represent all reads from one sample that overlap a region of interest in a
    reference sequence.

    Fields
    ------
    sample: str
        Name of the sample
    """

    def __init__(self, *, sample: str, **kwargs):
        super().__init__(**kwargs)
        self.sample = sample

    @property
    def path_fields(self):
        return {"sample": self.sample, **super().path_fields}

    @property
    def prof_fields(self):
        return {"sample": self.sample, **super().prof_fields}

    def __str__(self):
        return (f"Mutational Profile for sample '{self.sample}' "
                f"over reference '{self.ref}' region {self.end5}-{self.end3}")


class VectorIO(MutationalProfile):
    """
    Read and write mutation vectors.

    Fields
    ------
    top_dir: str
        Path to the top-level directory in which all temporary and final
        output files will be written
    """

    # Hashing algorithm to compute checksums of mutation vector files.
    digest_algo: ClassVar = "md5"

    def __init__(self, *, top_dir: path.TopDirPath, **kwargs):
        """
        Initialize a VectorIO object to read and write mutation vectors.
        
        Parameters
        ----------
        top_dir: str
        """
        super().__init__(**kwargs)
        self.top_dir = top_dir
        self.num_batches: int = 0
        self.num_vectors: int = 0
        self.checksums: list[str] = list()

    @property
    def path_fields(self):
        return {"top": self.top_dir.top, "partition": path.Partition.OUTPUT,
                "module": path.Module.VECTOR, **super().path_fields}

    @property
    def prof_fields(self):
        return {"top_dir": self.top_dir, "num_batches": self.num_batches,
                "num_vectors": self.num_vectors, "checksums": self.checksums,
                **super().prof_fields}

    @property
    def report_path(self):
        return path.MutVectorReportFilePath(**self.path_fields, ext=".json")

    @property
    def batch_dir(self):
        return path.RegionOutDirPath(**self.path_fields)

    def get_mv_batch_path(self, batch: int):
        return path.MutVectorBatchFilePath(**self.path_fields,
                                           batch=batch, ext=".orc")
    
    @property
    def batch_nums(self):
        """ List all the batch numbers. """
        return range(self.num_batches)

    @property
    def mv_batch_paths(self):
        return list(map(self.get_mv_batch_path, self.batch_nums))

    @classmethod
    def digest_file(cls, file_path: pathlib.Path) -> str:
        """
        Compute the checksum of a file.
        
        Parameters
        ----------
        file_path: Any
            Path of the file on which to compute the checksum. Can be any type
            as long as casting it to a string yields a valid path.
        
        Returns
        -------
        str
            Checksum of the file (in hexadecimal)
        """
        with open(file_path, "rb") as f:
            digest = file_digest(f, cls.digest_algo).hexdigest()
        return digest


class VectorReport(BaseModel, VectorIO):
    """
    Read and write a report about a mutational profile, with information such as
    - the sample, reference, and region
    - number of mutation vectors
    - number of mutation vector batch files and their checksums
    - beginning and ending time, duration, and speed of vectoring

    Fields
    ------
    top_str: str
        The top-level directory of the output files, as a string; writing to and
        reading from JSON format does not work with TopDirPath objects directly.
    sample: str
        Name of the sample
    

    Examples
    --------
    >>> report = VectorReport(top_str=os.getcwd(), sample="dms2",
    ...                       ref="tmv-rna", end5=1, end3=20,
    ...                       ref_str=DNA(b"GTATTTTTACAACAATTACC"),
    ...                       num_vectors=10346, num_batches=2,
    ...                       checksums=["b47260fcad8be60470bee67862f187b4",
    ...                                  "098f40cfadc266ea5bc48ab2e18cdc95"],
    ...                       began=datetime.now(),
    ...                       ended=(time.sleep(1E-5), datetime.now())[-1])
    >>> report.ref_str
    'GTATTTTTACAACAATTACC'
    """

    class Config:
        allow_population_by_field_name = True
        extra = Extra.ignore

    # Fields
    top_str: StrictStr = Field(alias="Top-level output directory")
    sample: StrictStr = Field(alias="Sample name")
    ref: StrictStr = Field(alias="Reference name")
    end5: PositiveInt = Field(alias="5' end of region")
    end3: PositiveInt = Field(alias="3' end of region")
    ref_str: StrictStr = Field(alias="Reference sequence")
    num_vectors: NonNegativeInt = Field(alias="Number of vectors")
    num_batches: NonNegativeInt = Field(alias="Number of batches")
    checksums: list[StrictStr] = Field(alias="MD5 checksums of vector batches")
    began: datetime = Field(alias="Began vectoring")
    ended: datetime = Field(alias="Ended vectoring")
    duration: NonNegativeFloat = Field(default=float("nan"),
                                       alias="Duration of vectoring (s)")
    speed: NonNegativeFloat = Field(default=float("nan"),
                                    alias="Speed of vectoring (vectors/s)")

    # Format of dates and times in the report file
    dt_fmt: ClassVar[str] = "on %Y-%m-%d at %H:%M:%S.%f"

    @validator("top_str", pre=True)
    def convert_top_dir_to_str(cls, top_str: str | path.TopDirPath):
        """ Return top-level directory (TopDirPath) as a string.
        Must be str in order to write and load from JSON correctly. """
        if isinstance(top_str, str):
            return top_str
        if isinstance(top_str, path.TopDirPath):
            return top_str.top
        raise TypeError(top_str)

    @property
    def top_dir(self):
        """ Return top-level directory string (from JSON) as a TopDirPath.
        The methods from VectorIO expect top_dir to be of type TopDirPath. """
        return path.TopDirPath.parse_path(self.top_str)

    @validator("ref_str", pre=True)
    def convert_ref_seq_to_str(cls, ref_str: DNA):
        """ Return reference sequence (DNA) as a string.
        Must be str in order to write to and load from JSON correctly. """
        return str(ref_str)

    @property
    def ref_seq(self):
        """ Return reference sequence string (from JSON) as a DNA object.
        The methods from VectorIO expect ref_seq to be of type DNA. """
        return DNA(self.ref_str.encode())

    @root_validator(pre=False)
    def calculate_duration_and_speed(cls, values):
        """
        Calculate and return the duration and speed of vectoring:
        - duration: difference between ending and beginning time (seconds)
        - speed: number of vectors processed per unit time (vectors/second)
        """
        began = values["began"]
        ended = values["ended"]
        dt = ended - began
        # Time delta objects store time as seconds and microseconds (both int).
        # These values must be converted to a float number of seconds like so:
        duration = dt.seconds + dt.microseconds / 1E6
        values["duration"] = duration
        if duration < 0:
            # Duration may not be negative.
            raise ValueError(f"Began {ended.strftime(cls.dt_fmt)}, but ended "
                             f"earlier, {began.strftime(cls.dt_fmt)}")
        num_vectors = values["num_vectors"]
        # Handle the unlikely case that duration == 0.0 by returning either
        # - inf (i.e. 1 / 0) if at least 1 vector was processed
        # - nan (i.e. 0 / 0) if no vectors were processed
        # when calculating the speed of processing vectors.
        speed = (float("inf" if num_vectors else "nan") if duration == 0.0
                 else round(num_vectors / duration, 1))
        values["speed"] = speed
        return values

    @root_validator(pre=False)
    def num_batches_len_checksums(cls, values):
        num_batches = values["num_batches"]
        num_checksums = len(values["checksums"])
        if num_batches != num_checksums:
            raise ValueError(f"Numbers of batches ({num_batches}) and "
                             f"checksums ({num_checksums}) did not match.")
        return values

    @root_validator(pre=False)
    def region_is_valid(cls, values):
        # The initialization of this Region instance will raise an error if the
        # region is not valid.
        Region(ref_seq=values["ref_str"], ref=values["ref"],
               end5=values["end5"], end3=values["end3"])
        return values

    def find_invalid_batches(self):
        """ Return all the batches of mutation vectors that either do not exist
        or do not match their expected checksums. """
        missing = list()
        badsum = list()
        for file, checksum in zip(self.mv_batch_paths, self.checksums,
                                  strict=True):
            try:
                if self.digest_file(file.pathstr) != checksum:
                    # The batch file exists but does not match the checksum.
                    badsum.append(file.pathstr)
            except FileNotFoundError:
                # The batch file does not exist.
                missing.append(file.pathstr)
        return missing, badsum

    def assert_valid_batches(self):
        missing, badsum = self.find_invalid_batches()
        if missing:
            raise FileNotFoundError(f"Missing vector batch files: {missing}")
        if badsum:
            raise ValueError(f"Batch files have bad checksums: {badsum}")

    def save(self):
        text = self.json(by_alias=True)
        self.report_path.path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.report_path.path, "w") as f:
            f.write(text)

    @classmethod
    def load(cls, file: Any):
        """ Load a mutation vector report from a file. """
        file_path = path.MutVectorReportFilePath.parse_path(file)
        report = cls.parse_file(file_path.path)
        if (file_path.top != report.top_dir.top
                or file_path.sample != report.sample
                or file_path.ref != report.ref
                or file_path.end5 != report.end5
                or file_path.end3 != report.end3):
            raise ValueError(f"Report fields do not match path '{file_path}'")
        report.assert_valid_batches()
        return report


class VectorWriter(VectorIO):
    """
    Compute mutation vectors for all reads from one sample mapping to one
    region of one reference sequence.
    """
    def __init__(self, *,
                 bam_path: path.OneRefAlignmentInFilePath,
                 min_qual: int,
                 rerun: bool,
                 **kwargs):
        super().__init__(sample=bam_path.sample,
                         ref=bam_path.ref,
                         **kwargs)
        self.bam_path = bam_path
        self.min_qual = min_qual
        self.seq_bytes = bytes(self.region_seq)
        self.rerun = rerun
        self._vectorized = False

    def _write_report(self, began: datetime, ended: datetime):
        report = VectorReport(**{**self.prof_fields,
                                 "top_str": self.top_dir,
                                 "ref_str": self.ref_seq,
                                 "began": began,
                                 "ended": ended})
        report.save()
        return report

    def _write_batch(self,
                     read_names: list[str, ...],
                     muts: tuple[bytearray, ...],
                     batch_num: int) -> tuple[path.MutVectorBatchFilePath, int]:
        """
        Write a batch of mutation vectors to an ORC file.

        ** Arguments **
        muts (NDArray) --> batch of mutation vectors in which each row is a
                           mutation vector and each column is a position in the
                           region of interest
        batch_num (int) -> non-negative integer label for the batch; every
                           mutational profile containing n batches includes all
                           batch numbers i in the range 0 <= i < n

        ** Returns **
        mv_file (str) <--- file path where the mutation vectors were written
        """
        # Convert the concatenated mutation vectors to a 2D NumPy array.
        muts_array = np.frombuffer(b"".join(muts), dtype=np.byte)
        n_records = len(muts_array) // self.length
        muts_array.resize((n_records, self.length))
        # Data must be converted to pd.DataFrame for PyArrow to write.
        # Explicitly set copy=False to copying the mutation vectors.
        df = pd.DataFrame(data=muts_array, index=read_names,
                          columns=self.columns, copy=False)
        mv_file = self.get_mv_batch_path(batch_num)
        df.to_orc(mv_file.path, engine="pyarrow")
        return mv_file, n_records

    def _vectorize_record(self, rec: SamRecord):
        """
        Compute the mutation vector of one record from a SAM file.

        ** Arguments **
        rec (SamRecord) --> SAM record for which to compute a mutation vector

        ** Returns **
        muts (bytearray) <- mutation vector
        """
        if rec.ref != self.ref:
            raise ValueError(f"Ref name of record ({rec.ref}) failed to "
                             f"match ref name of profile ({self.ref})")
        muts = rec.vectorize(self.seq_bytes, self.end5, self.end3)
        if not any(muts):
            raise ValueError("Mutation vector was entirely blank.")
        return rec.read_name, muts

    def _vectorize_batch(self, sam_viewer: SamViewer, batch_num: int,
                         start: int, stop: int):
        """
        Generate a batch of mutation vectors and write them to an ORC file.

        ** Arguments **
        sam_viewer (SamViewer) -> viewer to the SAM file for which to generate
                                  a batch of mutation vectors
        batch_num (int) --------> non-negative integer label for the batch
        start (int) ------------> seek to this position in the SAM file, then
                                  start generating vectors; position must be
                                  the beginning of a line, otherwise integrity
                                  checks will fail
        stop (int) -------------> stop generating vectors upon reaching this
                                  position in the SAM file; position must be
                                  the beginning of a line, otherwise integrity
                                  checks will fail
        
        ** Returns **
        n_records (int) <-------- number of records read from the SAM file
                                  between positions start and stop
        checksum (str) <--------- MD5 checksum of the ORC file of vectors
        """
        if stop > start:
            with sam_viewer as sv:
                # Use the SAM viewer to generate the mutation vectors.
                # Collect them as a single, 1-dimensional bytes object.
                read_names, muts = zip(*map(self._vectorize_record,
                                            sv.get_records(start, stop)))
        else:
            read_names, muts = (), ()
        # Write the mutation vectors to a file and compute its checksum.
        mv_file, n_records = self._write_batch(read_names, muts, batch_num)
        checksum = self.digest_file(mv_file.path)
        return n_records, checksum

    def _vectorize_sam(self, max_cpus: int):
        if self._vectorized:
            raise RuntimeError(f"Vectoring was already run for {self}.")
        self._vectorized = True
        with SamViewer(top_dir=self.top_dir,
                       max_cpus=max_cpus,
                       xam_path=self.bam_path,
                       ref_name=self.ref,
                       end5=self.end5,
                       end3=self.end3,
                       spanning=self.spanning,
                       min_qual=self.min_qual) as sv:
            batch_size = max(1, DEFAULT_BATCH_SIZE // self.length)
            indexes = list(sv.get_batch_indexes(batch_size))
            starts = indexes[:-1]
            stops = indexes[1:]
            self.num_batches = len(starts)
            svs = [SamViewer(top_dir=self.top_dir,
                             max_cpus=max_cpus,
                             xam_path=sv.sam_path,
                             ref_name=self.ref,
                             end5=self.end5,
                             end3=self.end3,
                             spanning=self.spanning,
                             min_qual=self.min_qual,
                             owner=False)
                   for _ in self.batch_nums]
            args = list(zip(svs, self.batch_nums, starts, stops, strict=True))
            if max_cpus > 1 and self.num_batches > 1:
                n_procs = min(max_cpus, self.num_batches)
                with Pool(n_procs, maxtasksperchild=1) as pool:
                    results = list(pool.starmap(self._vectorize_batch, args,
                                                chunksize=1))
            else:
                results = list(starmap(self._vectorize_batch, args))
            for num_vectors, checksum in results:
                self.num_vectors += num_vectors
                self.checksums.append(checksum)

    @property
    def outputs_valid(self):
        try:
            VectorReport.load(self.report_path.path)
        except (FileNotFoundError, ValueError):
            return False
        else:
            return True
    
    def vectorize(self, max_cpus: int):
        if self.rerun or not self.outputs_valid:
            self.batch_dir.path.mkdir(parents=True, exist_ok=True)
            logging.info(f"{self}: computing vectors")
            began = datetime.now()
            self._vectorize_sam(max_cpus)
            ended = datetime.now()
            logging.info(f"{self}: writing report")
            self._write_report(began, ended)
            logging.info(f"{self}: finished")
        else:
            logging.warning(f"{self}: already finished. To rerun, add --rerun")
        return self.report_path


class VectorWriterSpawner(object):
    def __init__(self,
                 top_dir: str,
                 fasta: str,
                 bam_files: list[str],
                 coords: list[tuple[str, int, int]],
                 primers: list[tuple[str, DNA, DNA]],
                 fill: bool,
                 parallel: str,
                 max_cpus: int,
                 min_phred: int,
                 phred_enc: int,
                 rerun: bool):
        self.top_dir = path.TopDirPath.parse_path(top_dir)
        self.bam_paths = [path.OneRefAlignmentInFilePath.parse_path(bam)
                          for bam in bam_files]
        self.ref_path = path.RefsetSeqInFilePath.parse_path(fasta)
        self.coords = coords
        self.primers = primers
        self.fill = fill
        self.parallel = parallel
        self.max_cpus = max_cpus
        self.min_phred = min_phred
        self.phred_enc = phred_enc
        self.rerun = rerun

    @property
    def min_qual(self):
        """
        Return the minimum quality for a base in a read to be considered
        informative, as the ASCII integer corresponding to the character in the
        FASTQ file that is the minimum valid quality.

        Returns
        -------
        int
            The ASCII value corresponding to the character in the FASTQ file read
            quality string that represents the minimum quality for a base to be
            considered informative.

        Examples
        --------
        For example, if the minimum Phred score (```self.min_phred```) that is
        accepted as informative is 20, and the Phred encoding of the FASTQ file
        (```phred_enc```) is 33 (i.e. ASCII+33), then the minimum quality as an
        ASCII integer (```min_qual```) is 20 + 33 = 53, which is character '5'.
        If ```min_phred``` were 37, then ```min_qual``` would be 37 + 33 = 70,
        which is character 'F'.
        """
        return self.min_phred + self.phred_enc
    
    @cached_property
    def bams_per_sample(self):
        samples: dict[str, list[
            path.OneRefAlignmentInFilePath]] = defaultdict(list)
        for bam in self.bam_paths:
            samples[bam.sample].append(bam)
        return samples
    
    @property
    def samples(self):
        return set(self.bams_per_sample.keys())
    
    @cached_property
    def ref_seqs(self):
        seqs = dict(FastaParser(self.ref_path.path).parse())
        if not seqs:
            raise ValueError(f"'{self.ref_path}' contained no sequences")
        return seqs
    
    @cached_property
    def regions(self):
        regions: dict[str, list[RegionFinder]] = defaultdict(list)

        def add_region(region: RegionFinder):
            if any(region == other for other in regions[region.ref]):
                raise ValueError(f"Duplicate region: {region.ref_coords}")
            regions[region.ref].append(region)

        for ref, first, last in self.coords:
            add_region(RegionFinder(ref_seq=self.ref_seqs.get(ref),
                                    ref=ref, end5=first, end3=last))
        for ref, fwd, rev in self.primers:
            add_region(RegionFinder(ref_seq=self.ref_seqs.get(ref),
                                    ref=ref, fwd=fwd, rev=rev))
        if self.fill:
            for ref, seq in self.ref_seqs.items():
                if ref not in regions:
                    add_region(RegionFinder(ref_seq=seq, ref=ref))
        return regions

    def _get_writers(self):
        for bam in self.bam_paths:
            for region in self.regions[bam.ref]:
                if region.ref != bam.ref:
                    raise RuntimeError(f"Refs of region ({region.ref}) "
                                       f"and BAM file ({bam.ref}) disagree.")
                yield VectorWriter(top_dir=self.top_dir,
                                   bam_path=bam,
                                   ref_seq=self.ref_seqs[bam.ref],
                                   end5=region.end5,
                                   end3=region.end3,
                                   min_qual=self.min_qual,
                                   rerun=self.rerun)

    @cached_property
    def writers(self):
        return list(self._get_writers())

    @property
    def num_writers(self):
        return len(self.writers)

    @property
    def parallel_broad(self):
        return (self.parallel == ParallelChoice.BROAD
                or (self.parallel == ParallelChoice.AUTO
                    and self.num_writers > 1))

    @property
    def parallel_deep(self):
        return (self.parallel == ParallelChoice.DEEP
                or (self.parallel == ParallelChoice.AUTO
                    and self.num_writers == 1))

    @cached_property
    def procs_per_profile(self):
        return self.max_cpus if self.parallel_deep else 1

    def generate_profiles(self):
        if self.num_writers == 0:
            raise ValueError("No BAM files and/or regions specified")
        args = tuple(zip(self.writers, repeat(self.procs_per_profile)))
        if self.parallel_broad:
            n_procs = max(min(self.max_cpus, self.num_writers), 1)
            with Pool(n_procs, maxtasksperchild=1) as pool:
                report_files = tuple(pool.starmap(VectorWriter.vectorize, args,
                                                  chunksize=1))
        else:
            report_files = tuple(starmap(VectorWriter.vectorize, args))
        return report_files


class VectorReader(VectorIO):
    @property
    def shape(self):
        return self.num_vectors, self.length

    @property
    def vectors(self):
        # FIXME: implement method to load the mutation vectors efficiently.
        return

    @classmethod
    def from_report(cls, report: VectorReport):
        return cls(top_dir=report.top_dir,
                   sample=report.sample,
                   ref_seq=report.ref_seq,
                   ref=report.ref,
                   end5=report.end5,
                   end3=report.end3)

    @classmethod
    def from_report_file(cls, report_file):
        return cls.from_report(VectorReport.load(report_file))



import itertools
import logging
from collections import defaultdict
from multiprocessing import Pool

from dreem.util.cli import DEFAULT_NEXTSEQ_TRIM, ParallelChoice
from dreem.util.seq import FastaParser, FastaWriter
from dreem.util.reads import (FastqAligner, FastqTrimmer, FastqUnit,
                              BamAlignSorter, BamSplitter,
                              SamRemoveEqualMappers)
from dreem.util.stargs import starstarmap
from dreem.util import path


def confirm_no_duplicate_samples(fq_units: list[FastqUnit]):
    # Count the number of times each sample and reference occurs.
    samples = defaultdict(int)
    sample_refs = defaultdict(lambda: defaultdict(int))
    for fq_unit in fq_units:
        if fq_unit.demult:
            sample_refs[fq_unit.sample][fq_unit.ref] += 1
        else:
            samples[fq_unit.sample] += 1
    # Find duplicates.
    dups = set()
    # Duplicate whole-sample FASTQs
    dups = dups | {sample for sample, count in samples.items() if count > 1}
    # Duplicate demultiplexed FASTQs
    dups = dups | {(sample, ref) for sample, refs in sample_refs.items()
                   for ref, count in refs.items() if count > 1}
    # Duplicate samples between whole-sample and demultiplexed FASTQs
    dups = dups | (set(samples) & set(sample_refs))
    if dups:
        # If there are any duplicate samples, raise an error.
        raise ValueError(f"Got duplicate samples/refs: {dups}")


def write_temp_ref_files(top_dir: path.TopDirPath,
                         refset_file: path.RefsetSeqInFilePath,
                         fq_units: list[FastqUnit]):
    ref_files: dict[str, path.OneRefSeqTempFilePath] = dict()
    # Determine which reference sequences need to be written.
    refs = {fq_unit.ref for fq_unit in fq_units if fq_unit.demult}
    if refs:
        # Only parse the FASTA if there are any references to write.
        for ref, seq in FastaParser(refset_file.path).parse():
            if ref in refs:
                ref_file = path.OneRefSeqTempFilePath(
                    top=top_dir.top,
                    partition=path.Partition.TEMP,
                    module=path.Module.ALIGN,
                    step=path.TempStep.ALIGN_ALIGN,
                    ref=ref,
                    ext=path.FASTA_EXTS[0])
                ref_path = ref_file.path
                ref_path.parent.mkdir(parents=True, exist_ok=True)
                logging.info(f"Writing temporary reference file: {ref_path}")
                FastaWriter(ref_path, {ref: seq}).write()
                ref_files[ref] = ref_file
    return ref_files


def run_steps_fq(top_dir: path.TopDirPath,
                 max_cpus: int,
                 fasta: path.RefsetSeqInFilePath | path.OneRefSeqTempFilePath,
                 fastq: FastqUnit,
                 nextseq_trim: bool = DEFAULT_NEXTSEQ_TRIM):
    # Trim the FASTQ file(s).
    trimmer = FastqTrimmer(top_dir, max_cpus, fastq)
    fastq = trimmer.run(nextseq_trim=nextseq_trim)
    # Align the FASTQ to the reference.
    aligner = FastqAligner(top_dir, max_cpus, fastq, fasta)
    xam_path = aligner.run()
    trimmer.clean()
    # Remove equally mapping reads.
    remover = SamRemoveEqualMappers(top_dir, max_cpus, xam_path)
    xam_path = remover.run()
    aligner.clean()
    # Sort the SAM file and output a BAM file.
    sorter = BamAlignSorter(top_dir, max_cpus, xam_path)
    xam_path = sorter.run()
    remover.clean()
    # Split the BAM file into one file for each reference.
    splitter = BamSplitter(top_dir, max_cpus, xam_path, fasta)
    bams = splitter.run()
    sorter.clean()
    return bams


def run_steps_fqs(top_path: str,
                  refset_path: str,
                  fq_units: list[FastqUnit],
                  parallel: str,
                  max_cpus: int,
                  **kwargs):
    if not fq_units:
        raise ValueError(f"No FASTQ files given")
    n_fqs = len(fq_units)  # > 0
    # Confirm that there are no duplicate samples.
    confirm_no_duplicate_samples(fq_units)
    # Generate the paths.
    top_dir = path.TopDirPath.parse_path(top_path)
    refset_file = path.RefsetSeqInFilePath.parse_path(refset_path)
    # Write the temporary FASTA files for demultiplexed FASTQs.
    ref_files = write_temp_ref_files(top_dir, refset_file, fq_units)
    try:
        # Determine if the computation should be parallelized
        # - broadly (i.e. run all FASTQs simultaneously in parallel and limit
        #   Cutadapt and Bowtie2 to single-threaded mode),
        # - deeply (i.e. run the FASTQs sequentially with Cutadapt and Bowtie2
        #   running in multithreaded mode),
        # - or with parallelization off (i.e. run the FASTQs sequentially and
        #   limit Cutadapt and Bowtie2 to single-threaded mode).
        parallel_broad = (parallel == ParallelChoice.BROAD
                          or (parallel == ParallelChoice.AUTO and n_fqs > 1))
        parallel_deep = (parallel == ParallelChoice.DEEP
                         or (parallel == ParallelChoice.AUTO and n_fqs == 1))
        # Compute the maximum number of processes allowed per FASTQ.
        procs_per_fq = max((max_cpus // n_fqs if parallel_broad
                            else max_cpus if parallel_deep
                            else 1), 1)
        # Next determine the positional and keyword arguments for each FASTQ.
        align_args = list()
        align_kwargs = list()
        for fq_unit in fq_units:
            # If the FASTQ is demultiplexed, then align to the FASTA containing
            # the one reference corresponding to the FASTQ; otherwise, align to
            # the oringal FASTA file that may contain more than one sequence.
            if fq_unit.demult:
                try:
                    fasta = ref_files[fq_unit.ref]
                except KeyError:
                    raise ValueError(f"Reference '{fq_unit.ref}' not found in "
                                     f"FASTA file {refset_file.path}")
            else:
                fasta = refset_file
            align_args.append((top_dir, procs_per_fq, fasta, fq_unit))
            align_kwargs.append(kwargs)
            print("FASTQ Unit", fq_unit, align_args[-1], align_kwargs[-1])
        if parallel_broad:
            # Process all FASTQs simultaneously in parallel.
            n_procs = max(min(len(fq_units), max_cpus), 1)
            with Pool(n_procs) as pool:
                bams = tuple(itertools.chain(*starstarmap(pool.starmap,
                                                          run_steps_fq,
                                                          align_args,
                                                          align_kwargs)))
        else:
            # Process each FASTQ file sequentially.
            bams = tuple(itertools.chain(*starstarmap(itertools.starmap,
                                                      run_steps_fq,
                                                      align_args,
                                                      align_kwargs)))
    finally:
        # Always delete the temporary files before exiting.
        for ref_file in ref_files.values():
            logging.info(f"Deleting temporary reference file: {ref_file.path}")
            ref_file.path.unlink(missing_ok=True)
    # Return a tuple of the final alignment map files.
    return bams



from pipeline import run
from dreem.util.cli import *


@click.command()
@optgroup.group('I/O')
@opt_top_dir
@arg_fasta
@opt_fastqs
@opt_fastqi
@opt_fastq1
@opt_fastq2
@opt_library
@opt_phred_enc
@opt_min_phred
@opt_rerun
@optgroup.group('Selection')
@opt_coords
@opt_primers
@opt_fill
@opt_parallel
@optgroup.group('Demultiplexing')
@demultiplexing
@barcode_start
@barcode_length
@max_barcode_mismatches
@optgroup.group('Clustering')
@clustering
@max_clusters
@min_iter
@signal_thresh
@info_thresh
@include_g_u
@include_del
@min_reads
@convergence_cutoff
@num_runs
@opt_cpus
@optgroup.group('Aggregation')
@rnastructure_path
@rnastructure_temperature
@rnastructure_fold_args
@rnastructure_dms
@rnastructure_dms_min_unpaired_value
@rnastructure_dms_max_paired_value
@rnastructure_partition
@rnastructure_probability


@optgroup.group('Misc')
@verbose
def cli(*args, **kwargs):
    run(*args, **kwargs)


if __name__ == '__main__':
    cli()



"""
Path module of DREEM
================================================================================
Version     0.0.2
Modified    2023-01-30
Author      Matty


Purpose
================================================================================
Several modules in DREEM produce files used by other modules.
For example, the alignment module creates alignment map files,
from which the vectoring module generates mutation vector files,
from which the aggregate module computes statistics and generates plots.

A module that writes a file and another module that reads the file must agree on
- the path to the file (so that the second module can find the file)
- the meaning of each part of the path (so that the second module can determine,
  e.g. the names of the sample and the reference from the path)

Although these path conventions could be written separately in each module,
this strategy is not ideal for several reasons:
- It would risk inconsistencies among the modules, potentially causing bugs.
- Changing the conventions would require modifying every module, which would
  not only be tedious but also increase the risk of inconsistencies.
- Defining all of the conventions in one place would reduce the size of the
  code base, improving readability, maintainability, and distribution.  

The path module defines the conventions for all file paths in one central
location, to be referenced by all other modules that use the file system.


Concepts
================================================================================

Paths and path segments
-----------------------
Every path is represented as a collection of path segments.
Each segment is the space between two path separators, except for the first
("top") segment of a path, which can contain separators.

For example, this path to an alignment map file
```/home/rfranklin/tmv/exp58/output/alignment/dms2/tmv-genome.bam```
would be represented as the following segments:
- top (full path of the top-level directory): ```/home/rfranklin/tmv/exp58/```
- partition (finished output or temporary files): ```output```
- module (DREEM module producing the results): ```alignment```
- sample (name of the sample from the experiment): ```dms2```
- ref (name of the reference sequence the sample was aligned to): tmv-genome

Classes of paths
----------------
(Description of what class you would use to represent several different kinds of paths)

Distinction between fields and 


Usage
================================================================================

Creating a path of a specific type using its class
--------------------------------------------------
An instance of a specific type of path can be created in three ways:

1. Calling the class directly, giving the names and values of the path segments
as keyword arguments:
>>> bam_path = OneRefAlignmentOutFilePath(top=os.getcwd(),
...                                       partition=Partition.OUTPUT,
...                                       module=Module.ALIGN, sample="dms2",
...                                       ref='tmv-rna', ext='.bam')
>>> assert str(bam_path.path) == (os.getcwd()
...                               + "/output/alignment/dms2/tmv-rna.bam")

2. Calling the class directly, giving the names and values of the path segments
as a dictionary of keyword arguments:
>>> bam_fields = {'top': os.getcwd(), 'partition': Partition.OUTPUT,
...               'module': Module.ALIGN, 'sample': 'dms2',
...               'ref': 'tmv-rna', 'ext': '.bam'}
>>> bam_path = OneRefAlignmentOutFilePath(**bam_fields)
>>> assert str(bam_path.path) == (os.getcwd()
...                               + "/output/alignment/dms2/tmv-rna.bam")

3. Parsing the path from a string (or from any other object whose __str__ method
returns a valid path, such as a pathlib.Path instance):
>>> path = os.path.join(os.getcwd(), "output/alignment/dms2/tmv-rna.bam")
>>> bam_path = OneRefAlignmentOutFilePath.parse_path(path)
>>> assert bam_path.dict() == {'top': os.getcwd(), 'partition': Partition.OUTPUT,
...                            'module': Module.ALIGN, 'sample': 'dms2',
...                            'ref': 'tmv-rna', 'ext': '.bam'}

Creating a path by inferring the type from the fields
-----------------------------------------------------
This module also provides a convenience function called make_path, which
behaves identically to the first two examples (of calling the class directly),
except that instead of the code specifying the type of path to be created,
make_path infers the type from the keywords arguments. For example:

A few notes on the function make_path:
- There is no equivalent of parse_path() for make_path because it is impossible
  to uniquely assign fields to segments of an arbitrary path if neither the
  structure nor the values of the fields are known.
- A TypeError is raised if there is no class of path whose field names match the
  names of the keyword arguments given to make_path.
- In most cases, this function suffices and alleviates the need to import every
  class of path that will be used in a module.


Implementation Details
================================================================================

Representation of paths and path segments
-----------------------------------------
In this module, every path is represented as a sequence of path segments. Every
kind of path is represented by a subclass of BasePath, and every kind of segment
is represented by a subclass of BaseSeg. For example, one subclass of BasePath
represents the full path to a mutation vector report file, and is made of path
segments that represent the name of the report file, the name of the directory
in which it resides, and so on up.

Responsibilities of paths and path segments
-------------------------------------------
Each path segment class is responsible for storing information contained in the
segment (i.e. the value of each field), generating a string representation of
the segment from the values of the fields, and parsing string representations
of the segment to determine the values of the fields encoded within. Each path
class is responsible for storing a series of path segments in the proper order,
ensuring that the path has all expected and no unexpected segments, ensuring
that the values of all of its segments are consistent with each other, and
generating and parsing string representations of the entire path by calling the
corresponding method of each segment and assembling the results.

Path segment classes and inheritance structure
----------------------------------------------
Every path segment class represents a specific component of a path, such as a
directory named after a sample or the name of a mutation vector batch file.

The base class of every path segment is BaseSeg, which does not represent any
specific path segment and should not be instantiated. Every other type of path
segment is a subclass of BaseSeg. BaseSeg is itself decorated with @dataclass
but contains no fields. However, it provides all shared methods for path segment
subclasses (all of which are also decorated with @dataclass):
- Initialization: __init__ (from @dataclass), __post_init__, validate
- Field access: fields, field_names, astuple, asdict, replace, _get_dtypes
- Interface: parse_seg, format_seg

The subclasses of BaseSeg that represent directories are structured as follows:

Subclass of BaseSeg     Fields          Constraints
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BaseSeg                 -               -
    TopSeg              top             directory must exist
    SubSeg              -               no field may contain a path separator
        PartSeg         partition       must be a valid partition name
        ModSeg          module          must be a valid module name
        TempStepSeg     step            must be a valid step name
        SampleSeg       sample          -
        RefsetSeg       refset          -
        RefSeg          ref             -
        StructSeg       -               -
            RegionSeg   end5, end3      1 ≤ end5 (int) ≤ end3 (int)
            ExtenSeg    ext             must be a valid extension for the class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Most segments for file types inherit from both ExtenSeg and another class that
provides one or more fields that are part of the file name:

Subclass of ExtenSeg    Also subclass of
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
MutVectorBatchSeg       -
MutVectorReportSeg	    RegionSeg
XamSeg                  -
    XamMixedSeg         RefsetSeg
    XamSplitSeg         RefSeg
FastaSeg	            -
    FastaMixedSeg	    RefsetSeg
    FastaSplitSeg	    RefSeg
FastqSeg                -
    FastqSampleSeg      SampleSeg
    FastqDemultSeg      RefSeg
Fastq1Seg               -
    Fastq1SampleSeg     SampleSeg
    Fastq1DemultSeg     RefSeg
Fastq2Seg               -
    Fastq2SampleSeg     SampleSeg
    Fastq2DemultSeg     RefSeg
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Implementation of path and path segment classes as Pydantic models
------------------------------------------------------------------


"""


# Imports ######################################################################

from __future__ import annotations
from enum import Enum
from functools import cache
from inspect import getmembers, isclass, signature
import itertools
import os
import pathlib
import re
from string import ascii_letters, digits
import sys
from typing import Any, ClassVar, Iterable

from pydantic import BaseModel, NonNegativeInt, PositiveInt, StrictStr, Extra
from pydantic import root_validator, validator


# Constants ####################################################################

# Valid/invalid characters in fields
VALID_CHARS = ascii_letters + digits + "_~=+-"
VALID_CHARS_SET = set(VALID_CHARS)
VALID_FIELD_PATTERN = f"([{VALID_CHARS}]+)"
VALID_FIELD_REGEX = re.compile(VALID_FIELD_PATTERN)


TOP_KEY = "top"
EXT_KEY = "ext"


class ValEnum(Enum):
    """ Subclass of Enum for which both str and repr return the value. """

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(str(self))


class Partition(ValEnum):
    OUTPUT = "output"
    TEMP = "temp"


class Module(ValEnum):
    DEMULT = "demultiplexing"
    ALIGN = "alignment"
    VECTOR = "vectoring"
    CLUSTER = "clustering"
    AGGREG = "aggregation"


class TempStep(ValEnum):
    ALIGN_TRIM = "align_1_trim"
    ALIGN_ALIGN = "align_2_align"
    ALIGN_REMEQ = "align_3_remeq"
    ALIGN_SORT = "align_4_sort"
    ALIGN_SPLIT = "align_5_split"
    VECTOR_SELECT = "vector_1_select"
    VECTOR_SORT = "vector_2_sort"


# File extensions
EXT_PATTERN = "([.].+)"
LOG_EXTS = (".log",)
FASTA_EXTS = (".fasta", ".fa")
FQ_EXTS = (".fastq", ".fq", ".fastq.gz", ".fq.gz")
FQ_PAIRED_EXTS_TEMPLATES = ("_R{}{}", "_mate{}{}", "_{}_sequence{}")
FQ1_EXTS = tuple(template.format(1, ext) for template, ext in
                 itertools.product(FQ_PAIRED_EXTS_TEMPLATES, FQ_EXTS))
FQ2_EXTS = tuple(template.format(2, ext) for template, ext in
                 itertools.product(FQ_PAIRED_EXTS_TEMPLATES, FQ_EXTS))
FQ1_PATTERNS = tuple(VALID_FIELD_PATTERN + f"({ext})" for ext in FQ1_EXTS)
FQ2_PATTERNS = tuple(VALID_FIELD_PATTERN + f"({ext})" for ext in FQ2_EXTS)
SAM_EXT = ".sam"
BAM_EXT = ".bam"
CRAM_EXT = ".cram"
XAM_EXTS = (SAM_EXT, BAM_EXT, CRAM_EXT)
BAI_EXT = f"{BAM_EXT}.bai"
XAI_EXTS = (BAI_EXT,)
XAMI_EXTS = XAM_EXTS + XAI_EXTS


# Path functions ###############################################################

def sanitize(path: Any):
    return os.path.realpath(os.path.normpath(os.path.abspath(str(path))))


# Path segment classes #########################################################

class BaseSeg(BaseModel):
    """ Abstract base class for all classes that represent segments of a path.
    Should not be instantiated. """

    class Config:
        extra = Extra.forbid

    @classmethod
    def fields(cls):
        return cls.__fields__

    @classmethod
    def n_fields(cls):
        return len(cls.fields())

    @classmethod
    def keys(cls):
        return list(cls.fields().keys())

    @classmethod
    def build_dict(cls, values: Iterable):
        return dict(zip(cls.keys(), values, strict=True))

    def values(self):
        return list(self.dict().values())

    @classmethod
    def parse_seg(cls, segment: Any):
        """ Return a new instance of the class given a segment as a str-like
        argument. Non-structured segments (including this base class) must have
        exactly one field. """
        return cls(**cls.build_dict([segment]))

    @property
    def seg_str(self):
        """ Return a string representation of the first field of the segment.
        Non-structured segments (including this base class) must have exactly
        one field, else a ValueError will be raised. """
        try:
            value, = self.values()
        except ValueError:
            raise ValueError(f"segstr is undefined for {self.__class__} "
                             f"with {self.n_fields()} (≠ 1) fields.")
        return str(value)


class TopSeg(BaseSeg):
    """ Class representing the top-level working directory of DREEM. All
    temporary and final output files will be located in this directory. This
    directory must exist at the time DREEM is launched. """
    top: StrictStr

    @validator(TOP_KEY)
    def top_dir_must_exist(cls, top):
        top_full = sanitize(top)
        if not os.path.isdir(top_full):
            raise NotADirectoryError(top_full)
        return top_full


class SubSeg(BaseSeg):
    @root_validator(pre=True)
    def fields_must_match_valid_field_regex(cls, values: dict[str, Any]):
        for key, value in values.items():
            if (key != TOP_KEY and key != EXT_KEY
                    and not VALID_FIELD_REGEX.match(str(value))):
                raise ValueError(f"{cls} got invalid '{key}' value: '{value}'")
        return values


class SampleSeg(SubSeg):
    """ Segment for a directory named after a sample. """
    sample: StrictStr


class RefsetSeg(SubSeg):
    """ Segment for a directory named after a set of reference sequences. """
    refset: StrictStr


class OneRefSeg(SubSeg):
    """ Segment for a directory named after one reference sequence. """
    ref: StrictStr


class PartSeg(SubSeg):
    """ Segment for a partition directory ('output' or 'temp'). """
    partition: Partition


class ModSeg(SubSeg):
    """ Segment for a directory containing the outputs of a module. """
    module: Module


class TempStepSeg(SubSeg):
    """ Segment for a directory containing the temporary files of a step. """
    step: TempStep


class BatchSeg(SubSeg):
    """ Segment for a batch. """
    batch: NonNegativeInt


class StructSeg(SubSeg):
    """
    Segment representing any path segment whose name contains an internal
    structure with one or more fields. For example, a segment could encode the
    name of a sample and the year, month, and day on which it was generated.
    This base class provides all the methods of structured segments but does
    not itself have any fields.

    Class attributes
    ----------------
    format_str: str
        Specify how to represent the path segment as a string, using Python's
        string formatting mini-language. The string is created by calling
        ```self.format_str.format(*self.dict())```.

        Note: ```format_str``` must be compatible with ```pattern_str```
    pattern_str: str
        Specify how to parse a string representing the path to determine the
        values of the fields. The string is parsed using ```re.match```,
        and ```match.groups()``` is used to get the values of the fields.

        Note: ```pattern_str``` must be compatible with ```format_str```

    Examples
    --------
    >>> class SampleBatchSeg(StructSeg, BatchSeg, SampleSeg):
    ...     format_str = "samp-{}_batch-{}"
    ...     pattern_str = f"samp-{VALID_FIELD_PATTERN}_batch-([0-9]+)"
    >>> seg = SampleBatchSeg(sample="tmv", batch=37)
    >>> seg.seg_str
    'samp-tmv_batch-37'
    """

    format_str: ClassVar[str] = ""
    pattern_str: ClassVar[str] = ""

    @classmethod
    def _parse_seg(cls, seg_str: Any):
        """ Return a new instance of the class by parsing a ```str```-like
        ```segstr``` with ```re.match``` (using ```pattern_str``` as the
        regular expression pattern), then passing the parsed groups to the
        class constructor. """
        # Try to parse segstr with re.match, using pattern_str as the pattern.
        if match := re.match(cls.pattern_str, str(seg_str)):
            # If segstr matched the pattern, then create keyword arguments from
            # the matched values, and use them to initialize a new instance.
            return cls(**cls.build_dict(match.groups()))
        # If segstr did not match the pattern, then the parsing failed.
        raise ValueError(
            f"Segment '{seg_str}' failed to match pattern {cls.pattern_str}")

    @classmethod
    def parse_seg(cls, segstr: Any):
        """
        Return a new instance of the class by parsing the values of the fields
        from a string and passing them to the class constructor.

        Parameters
        ----------
        segstr: Any
            Literal representation of the segment to parse as a ```str```
            (or as any other object ```x``` for which ```str(x)``` is a valid
            path segment, e.g. an instance of ```pathlib.Path```).
        
        Returns
        -------
        seginst: StructSeg
            New instance of a structured path segment with fields parsed from
            ```segstr```. Satisfies ```str(seginst) == str(segstr)```.

        Raises
        ------
        ValueError
            if the string representation of the newly created instance of the
            class did not match ```segstr```.
        """
        # Create a new instance of the class by parsing segstr.
        seginst = cls._parse_seg(segstr)
        # Confirm that formatting the new instance yields the original segstr.
        if (newstr := seginst._format_seg_str()) != str(segstr):
            # If not, an error occurred during parsing.
            raise ValueError(f"The new instance was formatted as '{newstr}' "
                             f"(failed to match input '{segstr}')")
        # If so, the new instance was parsed correctly.
        return seginst

    def _format_seg_str(self):
        """ Return the result of calling the ```format``` method of
        ```format_str``` with the field values as arguments. """
        return self.format_str.format(*self.values())

    @property
    def seg_str(self) -> str:
        """
        Return a string representation of the path segment by inserting the
        values of the segment's fields into its format specification.

        Returns
        -------
        str
            A representation of the path segment as a string.

        Raises
        ------
        ValueError
            if the resulting string cannot be parsed into a new instance whose
            fields match those of the current instance.
        """
        # Create the formatted string.
        segstr = self._format_seg_str()
        # Try to parse that string into a new instance, then confirm that the
        # fields of the new instance match those of the self instance.
        if (parse := self._parse_seg(segstr)) != self:
            # Raise an error if the fields did not match.
            raise ValueError(f"String representation '{segstr}' was parsed as"
                             f" {repr(parse)} (failed to match {repr(self)})")
        # If the fields match, then the formatted string is valid.
        return segstr


class RegionSeg(StructSeg):
    """
    Segment for a directory of a region of a reference sequence.

    Fields
    ------
    end5: int
        The 5'-most coordinate in the region; 1-indexed, inclusive.
    end3: int
        The 3'-most coordinate in the region; 1-indexed, inclusive.
    """

    end5: PositiveInt
    end3: PositiveInt

    format_str = "{}-{}"
    pattern_str = "([0-9]+)-([0-9]+)"

    @root_validator()
    def end5_le_end3(cls, values):
        """ Validate that end5 ≤ end3 """
        if values["end5"] > values["end3"]:
            raise ValueError(f"Got end5 ({values['end5']}) "
                             f"> end3 ({values['end3']})")
        return values


class FileSeg(StructSeg):
    """
    Segment representing a file name with an extension.

    Fields
    ------
    ext: str
        The file extension. Must begin with a period.

    Class attributes
    ----------------
    exts: tuple[str]
        All valid file extensions for the class. Each must begin with a period.
    """
    ext: str
    exts: ClassVar[tuple[str, ...]] = tuple()

    format_str = "{}{}"
    pattern_str = VALID_FIELD_PATTERN + EXT_PATTERN

    @validator(EXT_KEY)
    def valid_file_extension(cls, ext):
        """ Validate the file extension (```ext```). It must be an element of
        the class attribute ```exts```. """
        if ext not in cls.exts:
            raise ValueError(f"Invalid extension for {cls}: '{ext}'")
        return ext

    @classmethod
    def keys_no_ext(cls):
        """ Return a list of field keys without the 'ext' field. """
        # Create a new list of the field keys.
        keys = cls.keys()
        # Remove 'ext' from the list.
        keys.remove(EXT_KEY)
        # Return the list of keys without 'ext'.
        return keys

    @classmethod
    def _parse_seg(cls, segstr: Any):
        if all(ext[0] not in VALID_CHARS_SET for ext in cls.exts):
            # If the extension starts with a character that is not valid if it
            # occurs in a field (e.g. the most common starting character '.'),
            # then the extension can be distinguished from the other fields by
            # the regular expression defined in the superclass.
            return super()._parse_seg(segstr)
        # If the extension starts with a character that is valid if it occurs
        # in a field (e.g. '_R1.fq' is such an extension because it starts
        # with '_', which may occur in a field), then we need to match the
        # segment minus the extension to the pattern minus its extension.
        segstr_ = str(segstr)
        # Remove the final group, denoted by (), from the pattern string. This
        # group captures the extension, so it is removed.
        pattern_trunc = cls.pattern_str[:cls.pattern_str.rindex("(")]
        for ext in cls.exts:
            # Find the extension in exts that segstr ends with.
            if segstr_.endswith(ext):
                # Remove that extension from segstr.
                segstr_trunc = segstr_[:-len(ext)]
                # See if the truncated segment matches the truncated pattern.
                if match := re.match(pattern_trunc, segstr_trunc):
                    # If the truncated strings matched, then add the extension
                    # back to the end of the matched groups; we already know it
                    # matched because segstr_.endswith(ext) returned True.
                    return cls(**cls.build_dict(match.groups() + (ext,)))
        # If segstr did not match any extension, or did match an extension but
        # failed to match any patterns after removing the extension, then the
        # parsing failed.
        raise ValueError(f"Segment '{segstr}' failed to match pattern "
                         f"'{cls.pattern_str}' with any extension ({cls.exts})")


class MutVectorReportFileSeg(FileSeg, RegionSeg):
    """ Segment for a mutation vector report file. """
    format_str = "{}-{}_report{}"
    pattern_str = f"([0-9]+)-([0-9]+)_report{EXT_PATTERN}"
    exts = (".json",)


class MutVectorBatchFileSeg(FileSeg, BatchSeg):
    """ Segment for a mutation vector batch file. """
    format_str = "vectors_{}{}"
    pattern_str = "vectors_([0-9]+)" + EXT_PATTERN
    exts = (".orc",)


class AbstractRefFileSeg(FileSeg):
    """ Abstract segment for a FASTA file. Should not be instantiated. """
    exts = FASTA_EXTS


class RefsetSeqFileSeg(AbstractRefFileSeg, RefsetSeg):
    """ Segment for a FASTA file that can contain one or more references. The
    refset field is the name of the file, minus the file extension. """


class OneRefFileSeg(AbstractRefFileSeg, OneRefSeg):
    """ Segment for a FASTA file that contains exactly one reference sequence.
    The ref field is the reference name (on the first line, after '>'). """


class AbstractReadsFileSeg(FileSeg):
    """ Abstract segment for a standalone (single-end or paired-interleaved)
    FASTQ file. Should not be instantiated. """
    exts = FQ_EXTS


class SampleReadsFileSeg(AbstractReadsFileSeg, SampleSeg):
    """ Segment for a standalone FASTQ file containing all reads from a
    sample. """


class DemultReadsFileSeg(AbstractReadsFileSeg, OneRefSeg):
    """ Segment for a standalone FASTQ file containing all reads from one
    construct (i.e. reference sequence) in one sample. """


class AbstractReads1FileSeg(FileSeg):
    """ Abstract segment for a paired-end FASTQ file containing mate-1 reads.
    Should not be instantiated. """
    exts = FQ1_EXTS


class SampleReads1FileSeg(AbstractReads1FileSeg, SampleSeg):
    """ Segment for a paired-end FASTQ file containing all mate-1 reads from a
    sample. """


class DemultReads1FileSeg(AbstractReads1FileSeg, OneRefSeg):
    """ Segment for a paired-end FASTQ file containing all mate-1 reads from one
    construct (i.e. reference sequence) in one sample. """


class AbstractReads2FileSeg(FileSeg):
    """ Abstract segment for a paired-end FASTQ file containing mate-2 reads.
    Should not be instantiated. """
    exts = FQ2_EXTS


class SampleReads2FileSeg(AbstractReads2FileSeg, SampleSeg):
    """ Segment for a paired-end FASTQ file containing all mate-2 reads from a
    sample. """


class DemultReads2FileSeg(AbstractReads2FileSeg, OneRefSeg):
    """ Segment for a paired-end FASTQ file containing all mate-2 reads from one
    construct (i.e. reference sequence) in one sample. """


class AbstractAlignmentFileSeg(FileSeg):
    """ Segment representing both alignment map files (.sam, .bam, .cram)
    and index files (.bam.bai) """
    exts = XAMI_EXTS


class RefsetAlignmentFileSeg(AbstractAlignmentFileSeg, RefsetSeg):
    """ Segment representing an ailgnment map file resulting from aligning a
    set of reads (i.e. FASTQ) to a set of references (i.e. FASTA). """


class OneRefAlignmentFileSeg(AbstractAlignmentFileSeg, OneRefSeg):
    """ Segment representing an alignment map file produced by splitting another
    alignment map file into one file for each reference. """


class RegionAlignmentFileSeg(AbstractAlignmentFileSeg, RegionSeg):
    """ Segment representing an alignment map file produced by taking a subset
    of a particular region from another alignment map file. """
    format_str = "{}-{}{}"
    pattern_str = f"([0-9]+)-([0-9]+){EXT_PATTERN}"


# Path classes #################################################################


class BasePath(BaseModel):
    """ Abstract base class for all classes that represent full, absolute paths.
    Should not be instantiated. """

    @classmethod
    def _segment_types(cls):
        """ Return a list of the type of every segment in the path. """
        seg_types: list[type[BaseSeg]] = list()
        # Search the base classes for path segments in reverse order.
        for base in reversed(cls.__bases__):
            if issubclass(base, BasePath):
                # The base class represents a path and thus also has base
                # classes representing path segments: add them to the list.
                seg_types.extend(base._segment_types())
            elif issubclass(base, BaseSeg):
                # The base class is a path segment but not a path or path
                # scaffold class. Thus, it should be in the list of segments.
                seg_types.append(base)
            else:
                # Otherwise, the base class neither is a nor contains any
                # path segments. Skip it.
                pass
        return seg_types

    @classmethod
    @cache
    def segment_types(cls) -> tuple[type[BaseSeg], ...]:
        """ Return a tuple of the type of every segment in the path, in order
        from the beginning to the end of the path. """
        seg_types = tuple(cls._segment_types())
        if seg_types and seg_types[0] is not TopSeg:
            raise ValueError(f"{cls} begins with {seg_types[0]}, not {TopSeg}")
        return seg_types

    @property
    def segments(self) -> tuple[BaseSeg, ...]:
        """ Return a tuple of an instance of every segment in the path, in order
        from the beginning to the end of the path. """
        return tuple(seg_type(**{key: getattr(self, key)
                                 for key in seg_type.keys()})
                     for seg_type in self.segment_types())

    @classmethod
    def _parse_segments(cls, seg_types: list[type[BaseSeg]], path: str):
        """
        Return a dict of the names and values of the fields encoded within a
        string representation of a path.

        Parameters
        ----------
        seg_types: list[type[BaseSeg]]
            A list of the types of segments in the path, in order from beginning
            to end of the path. The first item must be ```TopSeg```.
        path: str
            The path to parse, represented as a string.

        Returns
        -------
        dict[str, Any]
            A dict of the names (keys) and values (values) of the fields encoded
            in the path string.

        Raises
        ------
        ValueError
            if any part of the path remains to be parsed after all the types of
            segments have been used.
        """
        if seg_types:
            # If at least one more segment type needs to be parsed, then get the
            # type that needs to be parsed now (from the end of seg_types).
            seg_type = seg_types.pop()
            if seg_types:
                # If any segment types need to be parsed after the current one
                # has been parsed, then parse the last segment of the path now
                # (with the current segment type), and parse everything above
                # it during the next recursive call (with the segment types
                # that need to be parsed after the current one).
                parse_next, parse_now = os.path.split(path)
            else:
                # If the current segment type is the last one that needs to be
                # parsed, then parse the entire remaining path with the current
                # segment type, and there is no part of the path to parse next.
                parse_next, parse_now = "", path
            # Parse the current segment with the current segment type, and the
            # next segment(s) with the remaining segment type(s), then merge.
            return {**cls._parse_segments(seg_types, parse_next),
                    **seg_type.parse_seg(parse_now).dict()}
        else:
            # No segment types still need to be parsed.
            if path:
                # Any part of the path that has not yet been parsed cannot be,
                # since no segment types are left to parse it.
                raise ValueError(f"No segments remain to parse '{path}'")
            # Return a dict with no fields, signifying that nothing remains to
            # be parsed.
            return {}

    @classmethod
    def parse_path(cls, path: Any):
        """
        Return a new instance of the class by parsing a given path.

        Parameters
        ----------
        path: Any
            A ```str``` representing the path to be parsed (or any other
            object ```x``` for which ```str(x)``` returns a valid path,
            e.g. an instance of ```pathlib.Path```).

        Returns
        -------
        BasePath
            A new instance of this class that represents the given path.

        Raises
        ------
        ValueError
            if the newly created instance of the class yields a string
            representation that does not match the ```path``` argument.
        """
        pathinst = cls(**cls._parse_segments(list(cls.segment_types()), path))
        if str(pathinst.path) != sanitize(path):
            raise ValueError(
                f"String representation of new path '{pathinst.path}' "
                f"failed to match input '{path}'")
        return pathinst

    @property
    def path(self):
        """ Return a ```pathlib.Path``` instance representing this path. """
        return pathlib.Path(*(segment.seg_str for segment in self.segments))

    @property
    def pathstr(self):
        """ Return a string representing this path. """
        return str(self.path)


# General directory paths

class TopDirPath(TopSeg, BasePath):
    def upto(self, field_name: str, include: bool = True):
        """
        Return a new instance of a path class representing the path up to the
        field ```field_name```.

        Parameters
        ----------
        field_name: str
            Name of the field up to which the new path should go.
        include: bool
            Whether to include the field ```field_name``` in the new path
            (default: ```True```).

        """
        keys_upto = (ks := self.keys())[:(n := ks.index(field_name) + include)]
        values_upto = self.values()[:n]
        return assemble(**dict(zip(keys_upto, values_upto)))
    
    def replace(self, **changes):
        """
        Return a new instance of a path class with some fields changed.

        Parameters
        ----------
        changes
            Keyword arguments giving the changes to make to the path. Each key
            is a field name and each value is the new value of the field.
        
        Returns
        -------
        TopDirPath
            A new instance of the same class with fields replaced by the values
            in ```changes```.
        """
        return self.__class__(**{**self.dict(), **changes})


class PartitionDirPath(PartSeg, TopDirPath):
    pass


class ModuleDirPath(ModSeg, PartitionDirPath):
    pass


class TempDirPath(TempStepSeg, ModuleDirPath):
    pass


# Sample, Reference, and Region directory paths

class SampleInDirPath(SampleSeg, TopDirPath):
    pass


class SampleTempDirPath(SampleSeg, TempDirPath):
    pass


class SampleOutDirPath(SampleSeg, ModuleDirPath):
    pass


class RefInDirPath(OneRefSeg, SampleInDirPath):
    pass


class RefTempDirPath(OneRefSeg, SampleTempDirPath):
    pass


class RefOutDirPath(OneRefSeg, SampleOutDirPath):
    pass


class RegionInDirPath(RegionSeg, RefInDirPath):
    pass


class RegionTempDirPath(RegionSeg, RefTempDirPath):
    pass


class RegionOutDirPath(RegionSeg, RefOutDirPath):
    pass


# Reference sequence (FASTA) file paths

class RefsetSeqInFilePath(RefsetSeqFileSeg, TopDirPath):
    pass


class OneRefSeqTempFilePath(OneRefFileSeg, TempDirPath):
    pass


# Read (FASTQ) file paths

class SampleReadsInFilePath(SampleReadsFileSeg, TopDirPath):
    pass


class SampleReads1InFilePath(SampleReads1FileSeg, TopDirPath):
    pass


class SampleReads2InFilePath(SampleReads2FileSeg, TopDirPath):
    pass


class SampleReadsTempFilePath(SampleReadsFileSeg, TempDirPath):
    pass


class SampleReads1TempFilePath(SampleReads1FileSeg, TempDirPath):
    pass


class SampleReads2TempFilePath(SampleReads2FileSeg, TempDirPath):
    pass


class SampleReadsOutFilePath(SampleReadsFileSeg, TempDirPath):
    pass


class SampleReads1OutFilePath(SampleReads1FileSeg, TempDirPath):
    pass


class SampleReads2OutFilePath(SampleReads2FileSeg, TempDirPath):
    pass


class DemultReadsInFilePath(DemultReadsFileSeg, SampleInDirPath):
    pass


class DemultReads1InFilePath(DemultReads1FileSeg, SampleInDirPath):
    pass


class DemultReads2InFilePath(DemultReads2FileSeg, SampleInDirPath):
    pass


class DemultReadsTempFilePath(DemultReadsFileSeg, SampleTempDirPath):
    pass


class DemultReads1TempFilePath(DemultReads1FileSeg, SampleTempDirPath):
    pass


class DemultReads2TempFilePath(DemultReads2FileSeg, SampleTempDirPath):
    pass


class DemultReadsOutFilePath(DemultReadsFileSeg, SampleOutDirPath):
    pass


class DemultReads1OutFilePath(DemultReads1FileSeg, SampleOutDirPath):
    pass


class DemultReads2OutFilePath(DemultReads2FileSeg, SampleOutDirPath):
    pass


# Alignment (SAM/BAM/CRAM) file paths

class RefsetAlignmentInFilePath(RefsetAlignmentFileSeg, SampleInDirPath):
    pass


class RefsetAlignmentTempFilePath(RefsetAlignmentFileSeg, SampleTempDirPath):
    pass


class RefsetAlignmentOutFilePath(RefsetAlignmentFileSeg, SampleTempDirPath):
    pass


class OneRefAlignmentInFilePath(OneRefAlignmentFileSeg, SampleInDirPath):
    pass


class OneRefAlignmentTempFilePath(OneRefAlignmentFileSeg, SampleTempDirPath):
    pass


class OneRefAlignmentOutFilePath(OneRefAlignmentFileSeg, SampleOutDirPath):
    pass


class RegionAlignmentInFilePath(RegionAlignmentFileSeg, RefInDirPath):
    pass


class RegionAlignmentTempFilePath(RegionAlignmentFileSeg, RefTempDirPath):
    pass


class RegionAlignmentOutFilePath(RegionAlignmentFileSeg, RefOutDirPath):
    pass


# Vectoring file paths

class MutVectorBatchFilePath(MutVectorBatchFileSeg, RegionOutDirPath):
    pass


class MutVectorReportFilePath(MutVectorReportFileSeg, RefOutDirPath):
    pass


# Path managing functions ######################################################

def is_path_class(query: Any):
    """
    Return whether ```query``` is a class of path that can be instantiated. It
    must be a subclass of both ```BasePath``` (to provide the methods for
    handling paths) and ```BaseSeg``` (to provide the segment(s) of the path,
    since every path that can be instantiated contains at least one segment).
    """
    return (isclass(query)
            and issubclass(query, BasePath)
            and issubclass(query, BaseSeg))


@cache
def _get_path_classes() -> dict[str, type[TopDirPath]]:
    """
    Return a ```dict``` of every valid, instantiable class of path defined in
    this module. Its keys are the class names (type ```str```) and its values
    the class objects (type ```type```).
    """
    return dict(getmembers(sys.modules[__name__], is_path_class))


def _sorted_fields(fields: Iterable[str]):
    """ Given an iterable of field names, return a tuple of the names in
    alphabetical order. """
    return tuple(sorted(fields))


@cache
def _get_path_signatures():
    return {name: _sorted_fields(signature(cls).parameters)
            for name, cls in _get_path_classes().items()}


def _path_class_from_signature(**fields):
    names = _sorted_fields(fields)
    matches = [_get_path_classes()[name]
               for name, sig in _get_path_signatures().items()
               if sig == names]
    if matches:
        if len(matches) > 1:
            raise TypeError(f"Multiple matching signatures for fields {names}")
        return matches[0]
    raise TypeError(f"No matching signature for fields: {names}")


def assemble(**fields):
    return _path_class_from_signature(**fields)(**fields)


# Path converters ##############################################################

class PathTypeTranslator(object):
    _trans: dict[type[TopDirPath], type[TopDirPath]] = dict()
    _inverse: type[PathTypeTranslator] | None = None

    @classmethod
    def items(cls):
        return cls._trans.items()

    @classmethod
    def trans_type(cls, input_type: type[TopDirPath]):
        """ Return the type to which a given type translates in the dictionary
        that this class defines. A thin wrapper around the dictionary's getitem
        method so that the dictionary does not need to be exposed. """
        try:
            return cls._trans[input_type]
        except KeyError:
            raise ValueError(f"{input_type} is not an input type for {cls}")

    @classmethod
    def trans_inst(cls, orig_inst: TopDirPath, preserve_type: bool = False,
                   **new_fields):
        """
        Translate an instance of a path of a given type to the output type to
        which it corresponds in this PathTypeMapper class. Optionally, add or
        modify the fields of the new type.

        Parameters
        ----------
        orig_inst: TopDirPath
            Instance of a path to be cast into a new type. The new type is
            determined by translating the type of orig_inst via the dictionary
            of types that this class of PathTypeMapper defines.
        preserve_type: bool = False
            Whether to return an instance of the same type as ```orig_inst```.
            If False (the default), the returned type will be the one to which
            the input type translates. If True, the output is converted back to
            the input type without changing the path that is represented by the
            output. Note that the output path cannot always be represented by
            the input type; if it cannot, a parsing error will be raised.
        **new_fields
            Keyword arguments of fields to add to or modify in the new instance.
            Any fields of the new type that are not in the original type must be
            given, else initializing the new instance will fail. To catch any
            mistyped field names, an error will be raised if any fields are not
            in the new type.

        Returns
        -------
        TopDirPath
            Instance of a path in the new type, possibly with new fields.

        Raises
        ------
        TypeError
            if any key of **new_fields is not a valid field in the output type.
        """
        # Determine the type of the instance to return.
        orig_type = type(orig_inst)
        new_type = cls.trans_type(orig_type)
        # Check if any given fields are not defined in the output type.
        new_keys = set(new_type.keys())
        if extras := set(new_fields) - new_keys:
            # If so, then in the best-case scenario, nothing happens; in the
            # worst-case scenario, a keyword was mistyped, potentially causing
            # a subtle bug. Raise an error to protect against the latter case.
            raise TypeError(f"Got keys not in {new_type}: {', '.join(extras)}")
        # Remove any fields that were defined in the input type but not in the
        # output type. This is not an error: any extra fields input type will
        # just get lumped into the catch-all 'top' field in the output type.
        orig_fields = {key: value for key, value in orig_inst.dict().items()
                       if key in new_keys}
        # Create the new instance by combining the original fields (minus any
        # that are not defined in the output type) with the new fields (if a
        # key appears in both original and new fields, the new value overrides
        # the original) and passing to the constructor of the new type.
        new_inst = new_type(**{**orig_fields, **new_fields})
        if preserve_type:
            # If the new path is to be cast back to the original data type,
            # then parse the new path using the original type. Note that it
            # could be impossible to parse the new path with the original
            # type, in which case an error will be raised during parsing.
            return orig_type.parse_path(new_inst.path)
        else:
            # Otherwise, return the new instance of the new type.
            return new_inst

    @classmethod
    def inverse(cls):
        """
        Return a new subclass of PathTypeMapper that translates the output
        types to the input types. Raise a TypeError if the translation is not
        invertible.
        """
        if cls._inverse is None:
            trans = dict()
            for type_in, type_out in cls.items():
                if type_out in trans:
                    raise TypeError(f"Translation of {cls} is not invertible")
                trans[type_out] = type_in

            class InverseTranslator(PathTypeTranslator):
                _trans = trans
                _inverse = cls

            InverseTranslator.__name__ = f"{cls.__name__}_InverseTranslator"
            cls._inverse = InverseTranslator

        return cls._inverse

    @staticmethod
    def chain(name: str,
              *translators: type[PathTypeTranslator]):
        """
        Return a new subclass of PathTypeMapper that chains together the given
        translators, passing the output of one into the next, and so on.
        """
        trans = dict()
        if len(translators) < 2:
            raise TypeError("At least two translators must be given.")
        for type_in, type_out in translators[0].items():
            for translator in translators[1:]:
                type_out = translator.trans_type(type_out)
            trans[type_in] = type_out

        class MapInToOut(PathTypeTranslator):
            _trans = trans

        MapInToOut.__name__ = name
        return MapInToOut

    @classmethod
    def compose(cls, other: type[PathTypeTranslator], name: str):
        return cls.chain(name, cls, other)


class ReadsInToReadsTemp(PathTypeTranslator):
    _trans = {SampleReadsInFilePath: SampleReadsTempFilePath,
              SampleReads1InFilePath: SampleReads1TempFilePath,
              SampleReads2InFilePath: SampleReads2TempFilePath,
              DemultReadsInFilePath: DemultReadsTempFilePath,
              DemultReads1InFilePath: DemultReads1TempFilePath,
              DemultReads2InFilePath: DemultReads2TempFilePath}


class ReadsTempToReadsOut(PathTypeTranslator):
    _trans = {SampleReadsTempFilePath: SampleReadsOutFilePath,
              SampleReads1TempFilePath: SampleReads1OutFilePath,
              SampleReads2TempFilePath: SampleReads2OutFilePath,
              DemultReadsTempFilePath: DemultReadsOutFilePath,
              DemultReads1TempFilePath: DemultReads1OutFilePath,
              DemultReads2TempFilePath: DemultReads2OutFilePath}


class ReadsInToAlignmentTemp(PathTypeTranslator):
    _trans = {SampleReadsInFilePath: RefsetAlignmentTempFilePath,
              SampleReads1InFilePath: RefsetAlignmentTempFilePath,
              SampleReads2InFilePath: RefsetAlignmentTempFilePath,
              DemultReadsInFilePath: OneRefAlignmentTempFilePath,
              DemultReads1InFilePath: OneRefAlignmentTempFilePath,
              DemultReads2InFilePath: OneRefAlignmentTempFilePath}


class AlignmentInToAlignmentTemp(PathTypeTranslator):
    _trans = {OneRefAlignmentInFilePath: OneRefAlignmentTempFilePath,
              RefsetAlignmentInFilePath: RefsetAlignmentTempFilePath}


class AlignmentTempToAlignmentOut(PathTypeTranslator):
    _trans = {OneRefAlignmentTempFilePath: OneRefAlignmentOutFilePath,
              RefsetAlignmentTempFilePath: RefsetAlignmentOutFilePath}


class AlignmentInToRegionAlignmentTemp(PathTypeTranslator):
    _trans = {OneRefAlignmentInFilePath: RegionAlignmentTempFilePath}


class AlignmentInToRegionAlignmentOut(PathTypeTranslator):
    _trans = {OneRefAlignmentInFilePath: RegionAlignmentOutFilePath}


ReadsInToReadsOut = ReadsInToReadsTemp.compose(
    ReadsTempToReadsOut, "ReadsInToReadsOut"
)


AlignmentInToAlignmentOut = AlignmentInToAlignmentTemp.compose(
    AlignmentTempToAlignmentOut, "AlignmentInToAlignmentOut")



import pathlib
from abc import ABC, abstractmethod
from enum import Enum, StrEnum
import itertools
import logging
import os
import re
from functools import cached_property
from typing import BinaryIO

from dreem.util import path
from dreem.util.cli import DEFAULT_LOCAL, DEFAULT_UNALIGNED, DEFAULT_DISCORDANT, DEFAULT_MIXED, DEFAULT_DOVETAIL, \
    DEFAULT_CONTAIN, DEFAULT_FRAG_LEN_MIN, DEFAULT_FRAG_LEN_MAX, DEFAULT_SEED_INTERVAL, \
    DEFAULT_GAP_BAR, DEFAULT_SEED_SIZE, DEFAULT_EXTENSIONS, DEFAULT_RESEED, DEFAULT_PADDING, \
    MATCH_BONUS, MISMATCH_PENALTY, N_PENALTY, REF_GAP_PENALTY, READ_GAP_PENALTY, MateOrientation
from dreem.util.cli import DEFAULT_MIN_BASE_QUALITY, DEFAULT_ILLUMINA_ADAPTER, DEFAULT_MIN_OVERLAP, DEFAULT_MAX_ERROR, \
    DEFAULT_INDELS, DEFAULT_NEXTSEQ_TRIM, DEFAULT_DISCARD_TRIMMED, DEFAULT_DISCARD_UNTRIMMED, DEFAULT_MIN_LENGTH, \
    DEFAULT_SCORE_MIN
from dreem.util.dflt import BUFFER_LENGTH
from dreem.util.excmd import FASTQC_CMD, CUTADAPT_CMD, BOWTIE2_CMD, \
    BOWTIE2_BUILD_CMD, SAMTOOLS_CMD, run_cmd
from dreem.util.seq import FastaParser

# General parameters
DEFAULT_INTERLEAVED = False
SAM_HEADER = b"@"
SAM_ALIGN_SCORE = b"AS:i:"
SAM_EXTRA_SCORE = b"XS:i:"
FASTQ_REC_LENGTH = 4
DEFAULT_MIN_MAPQ = 30

# FastQC parameters
DEFAULT_EXTRACT = False


class FastqUnit(object):
    """
    Unified interface for handling the following sets of sequencing reads:

    - One FASTQ file of single-end reads from one sample
    - One FASTQ file of interleaved, paired-end reads from one sample
    - Two FASTQ files of mate 1 and mate 2 paired-end reads from one sample
    - One FASTQ file of single-end reads originating from one reference sequence
      in one sample
    - One FASTQ file of interleaved, paired-end reads originating from one
      reference sequence in one sample
    - Two FASTQ files of mate 1 and mate 2 paired-end reads originating from one
      reference sequence in one sample

    """

    MAX_PHRED_ENC = 127  # 2^7 - 1

    class KeyName(StrEnum):
        SINGLE = "fastqs"  # single-end reads
        INTER = "fastqi"  # interleaved paired-end reads
        MATE1 = "fastq1"  # mate 1 paired-end reads
        MATE2 = "fastq2"  # mate 2 paired-end reads

    class Bowtie2Flag(StrEnum):
        SINGLE = "-U"
        INTER = "--interleaved"
        MATE1 = "-1"
        MATE2 = "-2"

    class KeySampleType(Enum):
        SINGLE = path.SampleReadsInFilePath
        INTER = path.SampleReadsInFilePath
        MATE1 = path.SampleReads1InFilePath
        MATE2 = path.SampleReads2InFilePath

    class KeyDemultType(Enum):
        SINGLE = path.DemultReadsInFilePath
        INTER = path.DemultReadsInFilePath
        MATE1 = path.DemultReads1InFilePath
        MATE2 = path.DemultReads2InFilePath

    class KeySampleDirName(StrEnum):
        SINGLE = "fastqs_dir"  # single-end reads
        INTER = "fastqi_dir"  # interleaved paired-end reads
        MATE12 = "fastq12_dir"  # mate 1 and mate 2 paired-end reads

    class KeySampleDirType(Enum):
        SINGLE = path.SampleInDirPath
        INTER = path.SampleInDirPath
        MATE12 = path.SampleInDirPath

    def __init__(self, *,
                 fastqs: (path.SampleReadsInFilePath |
                          path.DemultReadsInFilePath |
                          None) = None,
                 fastqi: (path.SampleReadsInFilePath |
                          path.DemultReadsInFilePath |
                          None) = None,
                 fastq1: (path.SampleReads1InFilePath |
                          path.DemultReads1InFilePath |
                          None) = None,
                 fastq2: (path.SampleReads2InFilePath |
                          path.DemultReads2InFilePath |
                          None) = None,
                 phred_enc: int):
        self._phred_enc = phred_enc
        self._init_inputs = dict(zip(self._get_keyname_strs(),
                                     (fastqs, fastqi, fastq1, fastq2),
                                     strict=True))
        self._validate_inputs()

    def _validate_inputs(self):
        _ = self.phred_arg
        _ = self.inputs
        _ = self.sample
        if self.demult:
            _ = self.ref


    @property
    def phred_enc(self):
        if 0 <= self._phred_enc <= self.MAX_PHRED_ENC:
            return self._phred_enc
        raise ValueError(self._phred_enc)

    @property
    def phred_arg(self):
        return f"--phred{self.phred_enc}"

    @classmethod
    def _get_keyname_strs(cls):
        return tuple(map(str, cls.KeyName))

    @classmethod
    def _test_single(cls, keys: tuple[str, ...]):
        """ Return whether the arguments match a single-end FASTQ.
        No validation is performed. """
        return keys == (cls.KeyName.SINGLE,)

    @classmethod
    def _test_interleaved(cls, keys: tuple[str, ...]):
        """ Return whether the arguments match an interleaved FASTQ.
        No validation is performed. """
        return keys == (cls.KeyName.INTER,)

    @classmethod
    def _test_separate_mates(cls, keys: tuple[str, ...]):
        """ Return whether the arguments match two separate paired-end FASTQs.
        No validation is performed. """
        return keys == (cls.KeyName.MATE1, cls.KeyName.MATE2)

    @cached_property
    def _input_keys(self):
        """ Validate that a correct set of keyword arguments was given, and
        then return them. """
        # Find all the keys of self._inputs that got a value (not None).
        keys = tuple(key for key in self._get_keyname_strs()
                     if self._init_inputs[key] is not None)
        # The keys must match one of the following patterns to be valid.
        if not any((self._test_single(keys),
                    self._test_interleaved(keys),
                    self._test_separate_mates(keys))):
            raise ValueError(f"Got invalid FASTQ keywords: {keys}")
        # Now the keys are validated, and can validate other properties.
        return keys

    @property
    def _inputs(self):
        """ Return a dictionary of keyword arguments and path instances of the
        input FASTQ files. Keys are validated, but values and types are not. """
        return {key: self._init_inputs[key] for key in self._input_keys}

    @property
    def single(self):
        """ Return whether the arguments match a single-end FASTQ.
        Validated. """
        return self._test_single(self._input_keys)

    @property
    def interleaved(self):
        """ Return whether the arguments match an interleaved FASTQ.
        Validated. """
        return self._test_interleaved(self._input_keys)

    @property
    def separate_mates(self):
        """ Return whether the arguments match two separate paired-end FASTQs.
        Validated. """
        return self._test_separate_mates(self._input_keys)

    @property
    def paired(self):
        """ Return whether the reads in the FASTQ unit are paired-end.
        Validated. """
        return not self.single

    @classmethod
    def _test_by_sample(cls, keys: tuple[str, ...], types: tuple[type, ...]):
        """ Return whether the given path types match the path types expected
        for FASTQ files representing an entire sample. """
        return types == tuple(cls.KeySampleType[cls.KeyName(key).name].value
                              for key in keys)

    @classmethod
    def _test_by_ref(cls, keys: tuple[str, ...], types: tuple[type, ...]):
        """ Return whether the given path types match the path types expected
        for FASTQ files representing one reference from a sample. """
        return types == tuple(cls.KeyDemultType[cls.KeyName(key).name].value
                              for key in keys)

    @cached_property
    def _input_types(self):
        """ Validate that the types of paths given as input arguments match the
        expected types. """
        # Get the types of the input arguments.
        types = tuple(map(type, self._inputs.values()))
        # The types must match one of the expected set of path types given the
        # keyword arguments -- either the sample- or ref-based paths.
        if (self._test_by_sample(self._input_keys, types)
                == self._test_by_ref(self._input_keys, types)):
            raise TypeError(f"Got invalid FASTQ types: {types}")
        # Now the types are validated and can validate other properties.
        return types

    @cached_property
    def inputs(self):
        """ Return a dictionary of keyword arguments and path instances of the
        input FASTQ files. Validated. """
        inputs: dict[str, path.SampleReadsInFilePath |
                          path.SampleReads1InFilePath |
                          path.SampleReads2InFilePath |
                          path.DemultReadsInFilePath |
                          path.DemultReads1InFilePath |
                          path.DemultReads2InFilePath] = dict()
        for (key, path_inst), exp_type in zip(self._inputs.items(),
                                              self._input_types,
                                              strict=True):
            # Confirm that every item of self._inputs has the correct type.
            if (path_type := type(path_inst)) is not exp_type:
                raise TypeError(
                    f"Expected {exp_type} for key '{key}', but got {path_type}")
            inputs[key] = path_inst
        # Now the inputs are validated.
        return inputs

    @property
    def demult(self):
        return self._test_by_ref(self._input_keys, self._input_types)

    @property
    def paths(self):
        return tuple(inp.path for inp in self.inputs.values())

    @cached_property
    def sample(self):
        samples: tuple[str] = tuple({fq.sample for fq in self.inputs.values()})
        if len(samples) != 1:
            raise ValueError(f"Required exactly one sample, but got {samples}")
        return samples[0]

    @cached_property
    def ref(self):
        if not self.demult:
            raise TypeError("Whole-sample FASTQ files do not define reference")
        refs: tuple[str] = tuple({fq.ref for fq in self.inputs.values()})
        if len(refs) != 1:
            raise ValueError(f"Required exactly one reference, but got {refs}")
        return refs[0]

    @property
    def cutadapt_input_args(self):
        return self.paths

    @property
    def _bowtie2_flags(self):
        return tuple(str(self.Bowtie2Flag[self.KeyName(key).name])
                     for key in self._input_keys)

    @property
    def bowtie2_inputs(self):
        return tuple(itertools.chain(*map(list, zip(self._bowtie2_flags,
                                                    self.paths,
                                                    strict=True))))

    def trans(self, trans: type[path.PathTypeTranslator], **kwargs):
        """
        Return a new FastqUnit by translating all the FASTQ paths using the
        given PathTypeTranslator and optionally extra arguments.

        Parameters
        ----------
        trans: PathTypeTranslator
            Translation from one FASTQ type to another
        **kwargs: Any
            Keyword arguments passed to the ```trans_inst``` method of the
            path type translator

        Returns
        -------
        FastqUnit
            New FastqUnit with the same keys and translated paths
        """
        return FastqUnit(**{key: trans.trans_inst(inp, **kwargs)
                            for key, inp in self.inputs.items()},
                         phred_enc=self.phred_enc)

    @classmethod
    def _get_demult_files(cls, phred_enc: int, dir_path: path.BasePath,
                          key: str):
        """
        Yield a FastqUnit for each demultiplexed FASTQ file in a directory.

        Parameters
        ----------
        phred_enc: int
            Phred score encoding
        dir_path: SampleInDirPath
            Directory containing the FASTQ files
        key: str
            Keyword argument for every FASTQ file ('fastqs' or 'fastqi')

        Return
        ------
        Iterable[FastqUnit]
            One for each FASTQ file in the directory
        """
        dp = dir_path.path
        parse_type = cls.KeyDemultType[cls.KeyName(key).name].value
        for fname in dp.iterdir():
            yield cls(**{key: parse_type.parse_path(dp.joinpath(fname))},
                      phred_enc=phred_enc)

    @classmethod
    def _get_demult_pairs(cls, phred_enc: int, dir_path: path.BasePath):
        """
        Yield a FastqUnit for each pair of mated, demultiplexed FASTQ files
        in a directory.

        Parameters
        ----------
        phred_enc: int
            Phred score encoding
        dir_path: SampleInDirPath
            Directory containing the FASTQ files

        Return
        ------
        Iterable[FastqUnit]
            One for each FASTQ pair in the directory
        """
        dp = dir_path.path
        # Create empty dictionaries to store the FASTQ files for mates 1 and 2,
        # keyed by the name of the reference sequence for each FASTQ file.
        mates1: dict[str, path.DemultReads1InFilePath] = dict()
        mates2: dict[str, path.DemultReads2InFilePath] = dict()
        # Read the name of every file in the directory.
        for fname in dp.iterdir():
            # Get the full path to the file.
            fpath = str(dp.joinpath(fname))
            # Determine if the file is a mate 1 or mate 2 FASTQ file.
            is1 = any(fpath.endswith(ext) for ext in path.FQ1_EXTS)
            is2 = any(fpath.endswith(ext) for ext in path.FQ2_EXTS)
            if is1 and is2:
                # There should be no way for this error to happen, but catching
                # it just in case.
                raise ValueError(f"FASTQ path matched both mates: '{fpath}'")
            if is1:
                # The file name matched an extension for a mate 1 FASTQ file.
                fq = path.DemultReads1InFilePath.parse_path(fpath)
                if fq.ref in mates1:
                    raise ValueError(f"Got >1 FASTQ file for ref '{fq.ref}'")
                # Add the path to the dict of mate 1 files, keyed by reference.
                mates1[fq.ref] = fq
            elif is2:
                # The file name matched an extension for a mate 2 FASTQ file.
                fq = path.DemultReads2InFilePath.parse_path(fpath)
                if fq.ref in mates2:
                    raise ValueError(f"Got >1 FASTQ file for ref '{fq.ref}'")
                # Add the path to the dict of mate 2 files, keyed by reference.
                mates2[fq.ref] = fq
            else:
                # If a file name does not match the expected FASTQ name format,
                # log a warning but keep going, since the presence or absence
                # of one FASTQ file will not affect the others, and this file
                # might just be an extraneous file, such as .DS_Store on macOS.
                logging.warning(f"File name is not a valid FASTQ: '{fpath}'")
        # Determine if any reference names are present for either mate 1 or
        # mate 2 but not both mates.
        refs1 = set(mates1)
        refs2 = set(mates2)
        # Find union of all reference names from both mate 1 and mate 2 files.
        refs = refs1 | refs2
        # Find any reference names that exist for one mate but not both.
        if missing := refs - (refs1 & refs2):
            # If any reference names are missing for either mate, this is an
            # error because the FASTQ files should be generated in pairs, and
            # if not that indicates a problem, such as the accidental deletion
            # of one or more FASTQ files in the directory or the directory
            # actually containing single-end or interleaved FASTQ files.
            raise ValueError(f"Missing mate 1/2 for refs {', '.join(missing)}")
        for ref in refs:
            # Yield a FastqUnit for each pair of mated FASTQ files.
            yield cls(fastq1=mates1[ref], fastq2=mates2[ref],
                      phred_enc=phred_enc)

    @classmethod
    def _get_sample_files(cls, phred_enc: int, path_strs: tuple[str], key: str):
        file_type = cls.KeySampleType[cls.KeyName(key).name].value
        for path_str in path_strs:
            yield FastqUnit(**{key: file_type.parse_path(path_str)},
                            phred_enc=phred_enc)

    @classmethod
    def _get_sample_pairs(cls, phred_enc: int,
                          path1_strs: tuple[str],
                          path2_strs: tuple[str]):
        keys = (cls.KeyName.MATE1, cls.KeyName.MATE2)
        for path_strs in zip(path1_strs, path2_strs, strict=True):
            paths = {key.value:
                         cls.KeySampleType[key.name].value.parse_path(path_str)
                     for key, path_str in zip(keys, path_strs, strict=True)}
            yield FastqUnit(**paths, phred_enc=phred_enc)

    @classmethod
    def from_strs(cls, *, phred_enc: int, **fastq_args: tuple[str]):
        path1_strs = ()
        for fq_key, path_strs in fastq_args.items():
            try:
                # Check if fq_key is a key for a directory or file.
                dir_type = cls.KeySampleDirType[
                    cls.KeySampleDirName(fq_key).name].value
            except ValueError:
                # fq_key is not a key for a directory: assume it is a file key.
                if fq_key == cls.KeyName.MATE1 or fq_key == cls.KeyName.MATE2:
                    if not path1_strs:
                        try:
                            path1_strs = fastq_args[cls.KeyName.MATE1]
                            path2_strs = fastq_args[cls.KeyName.MATE2]
                        except KeyError:
                            raise KeyError(
                                f"Must give both {cls.KeyName.MATE1} and "
                                f"{cls.KeyName.MATE2} or neither, not only one")
                        yield from cls._get_sample_pairs(phred_enc=phred_enc,
                                                         path1_strs=path1_strs,
                                                         path2_strs=path2_strs)
                else:
                    yield from cls._get_sample_files(phred_enc=phred_enc,
                                                     path_strs=path_strs,
                                                     key=fq_key)
            else:
                # The FASTQ key is a key for a directory.
                # For each path string, parse the directory path.
                for dir_path in map(dir_type.parse_path, path_strs):
                    # Yield each FastqUnit depending on the FASTQ key.
                    if fq_key == cls.KeySampleDirName.MATE12:
                        # The key indicates that the directory contains paired
                        # FASTQ files split into mate 1 and mate 2 reads.
                        yield from cls._get_demult_pairs(phred_enc=phred_enc,
                                                         dir_path=dir_path)
                    else:
                        # The key indicates that the directory contains FASTQ
                        # files that may be processed individually.
                        yield from cls._get_demult_files(phred_enc=phred_enc,
                                                         dir_path=dir_path,
                                                         key=fq_key)


class ReadsFileBase(ABC):
    partition = path.Partition.TEMP
    module = path.Module.ALIGN
    step = ""
    ext = ""

    def __init__(self, top_dir: path.TopDirPath, max_cpus: int):
        self.top_dir = top_dir
        self.max_cpus = max_cpus

    @property
    @abstractmethod
    def sample(self):
        raise NotImplementedError

    @property
    def demult(self):
        raise NotImplementedError

    @property
    def fields(self):
        fields = {"top": self.top_dir.top,
                  "partition": self.partition,
                  "module": self.module,
                  "sample": self.sample}
        if self.partition == path.Partition.TEMP:
            fields["step"] = self.step
        return fields

    @cached_property
    def output_dir(self):
        if self.partition == path.Partition.TEMP:
            return path.SampleTempDirPath(**self.fields)
        if self.partition == path.Partition.OUTPUT:
            return path.SampleOutDirPath(**self.fields)
        raise ValueError(self.partition)

    @cached_property
    @abstractmethod
    def output(self):
        raise NotImplementedError

    def setup(self):
        self.output_dir.path.mkdir(parents=True, exist_ok=True)

    @abstractmethod
    def run(self):
        raise NotImplementedError

    @abstractmethod
    def clean(self):
        raise NotImplementedError


class FastqBase(ReadsFileBase):
    def __init__(self,
                 top_dir: path.TopDirPath,
                 max_cpus: int,
                 fastq: FastqUnit):
        super().__init__(top_dir, max_cpus)
        self.fastq = fastq

    @property
    def sample(self):
        return self.fastq.sample

    @property
    def demult(self):
        return self.fastq.demult

    def qc(self, extract: bool = DEFAULT_EXTRACT):
        cmd = [FASTQC_CMD]
        if extract:
            cmd.append("--extract")
        cmd.extend(self.fastq.paths)
        run_cmd(cmd)


class FastqTrimmer(FastqBase):
    step = path.TempStep.ALIGN_TRIM

    @cached_property
    def output(self):
        return self.fastq.trans(path.ReadsInToReadsTemp,
                                **self.fields,
                                preserve_type=True)

    @property
    def _cutadapt_output_flags(self):
        return "-o", "-p"

    @property
    def _cutadapt_output_args(self):
        return tuple(itertools.chain(*zip(self._cutadapt_output_flags,
                                          self.output.paths,
                                          strict=False)))

    def _cutadapt(self,
                  qual1: int = DEFAULT_MIN_BASE_QUALITY,
                  qual2: int = 0,
                  adapters15: tuple[str] = (),
                  adapters13: tuple[str] = (DEFAULT_ILLUMINA_ADAPTER,),
                  adapters25: tuple[str] = (),
                  adapters23: tuple[str] = (DEFAULT_ILLUMINA_ADAPTER,),
                  min_overlap: int = DEFAULT_MIN_OVERLAP,
                  max_error: float = DEFAULT_MAX_ERROR,
                  indels: bool = DEFAULT_INDELS,
                  nextseq_trim: bool = DEFAULT_NEXTSEQ_TRIM,
                  discard_trimmed: bool = DEFAULT_DISCARD_TRIMMED,
                  discard_untrimmed: bool = DEFAULT_DISCARD_UNTRIMMED,
                  min_length: bool = DEFAULT_MIN_LENGTH):
        cmd = [CUTADAPT_CMD]
        cmd.extend(["--cores", self.max_cpus])
        if nextseq_trim:
            if qual1 > 0:
                cmd.extend(["--nextseq-trim", qual1])
        else:
            if qual1 > 0:
                cmd.extend(["-q", qual1])
            if qual2 > 0:
                cmd.extend(["-Q", qual2])
        adapters = {"g": adapters15, "a": adapters13,
                    "G": adapters25, "A": adapters23}
        for arg, adapter in adapters.items():
            if adapter and (self.fastq.paired or arg.islower()):
                for adapt in adapter:
                    cmd.extend([f"-{arg}", adapt])
        if min_overlap >= 0:
            cmd.extend(["-O", min_overlap])
        if max_error >= 0:
            cmd.extend(["-e", max_error])
        if not indels:
            cmd.append("--no-indels")
        if discard_trimmed:
            cmd.append("--discard-trimmed")
        if discard_untrimmed:
            cmd.append("--discard-untrimmed")
        if min_length:
            cmd.extend(["-m", min_length])
        cmd.extend(["--report", "minimal"])
        if self.fastq.interleaved:
            cmd.append("--interleaved")
        cmd.extend(self._cutadapt_output_args)
        cmd.extend(self.fastq.cutadapt_input_args)
        run_cmd(cmd)
        return self.output

    def run(self, **kwargs):
        self.setup()
        return self._cutadapt(**kwargs)

    def clean(self):
        for out_path in self.output.paths:
            out_path.unlink()


class FastqAligner(FastqBase):
    step = path.TempStep.ALIGN_ALIGN
    ext = path.SAM_EXT

    def __init__(self,
                 top_dir: path.TopDirPath,
                 max_cpus: int,
                 fastq: FastqUnit,
                 fasta: path.RefsetSeqInFilePath | path.OneRefSeqTempFilePath):
        super().__init__(top_dir, max_cpus, fastq)
        if isinstance(fasta, path.RefsetSeqInFilePath):
            if self.demult:
                raise TypeError("Got a multi-FASTA but a demultiplexed FASTQ")
        elif isinstance(fasta, path.OneRefSeqTempFilePath):
            if not self.demult:
                raise TypeError("Got a single-FASTA but a multiplexed FASTQ")
        else:
            raise TypeError(fasta)
        self.fasta = fasta

    @property
    def fasta_prefix(self):
        return self.fasta.path.with_suffix("")

    @cached_property
    def output(self):
        # fasta provides either 'refset' or 'ref'
        # output_dir provides 'top', 'partition', 'module', 'step', and 'sample'
        # ext provides the file extension
        fields = {**self.fasta.dict(),
                  **self.fields,
                  path.EXT_KEY: self.ext}
        outputs = list()
        for fq in self.fastq.inputs.values():
            temp_inst = path.ReadsInToAlignmentTemp.trans_inst(fq, **fields)
            in_type = path.AlignmentInToAlignmentTemp.inverse().trans_type(
                type(temp_inst))
            in_inst = in_type.parse_path(temp_inst.path)
            outputs.append(in_inst)
        if not outputs:
            raise ValueError("No output files")
        if any(output != outputs[0] for output in outputs[1:]):
            raise ValueError("Inconsistent output files")
        return outputs[0]

    def _bowtie2_build(self):
        """ Build an index of a reference genome using Bowtie 2. """
        cmd = [BOWTIE2_BUILD_CMD, "-q", self.fasta.path, self.fasta_prefix]
        run_cmd(cmd)

    def _bowtie2(self,
                 local=DEFAULT_LOCAL,
                 unaligned=DEFAULT_UNALIGNED,
                 discordant=DEFAULT_DISCORDANT,
                 mixed=DEFAULT_MIXED,
                 dovetail=DEFAULT_DOVETAIL,
                 contain=DEFAULT_CONTAIN,
                 score_min=DEFAULT_SCORE_MIN,
                 frag_len_min=DEFAULT_FRAG_LEN_MIN,
                 frag_len_max=DEFAULT_FRAG_LEN_MAX,
                 gap_bar=DEFAULT_GAP_BAR,
                 seed_size=DEFAULT_SEED_SIZE,
                 seed_interval=DEFAULT_SEED_INTERVAL,
                 extensions=DEFAULT_EXTENSIONS,
                 reseed=DEFAULT_RESEED,
                 padding=DEFAULT_PADDING,
                 orientation=MateOrientation.FR):
        """
        Run alignment with Bowtie 2 (command ```bowtie2```).

        Parameters
        ----------
        local: bool
            Whether to find local (True) or end-to-end (False) read alignments
        gap_bar: int
            Forbid gaps within this many positions from the end of a read
        padding: int
            Number of positions to pad on each side of the matrix to allow gaps
        seed_size: int
            Length of the alignment seeds during multiseed alignment
        seed_interval: str (function)
            Interval between alignment seeds during multiseed alignment

        extensions: int
            Maximum number of failed seed extensions before moving on
        reseed: int
            Maximum number of attempts to re-seed repetitive seeds

        score_min: str (function)
            Minimum score to consider an alignment "good enough" to output
        frag_len_min: int (paired-end reads only)
            Minimum length of the contiguous fragment containing both mates
        frag_len_max: int (paired-end reads only)
            Maximum length of the contiguous fragment containing both mates
        unaligned: bool
            Whether to output reads that do not align

        orientation: str (paired-end reads only)
            Upstream/downstream mate orientations for a valid alignment
        discordant: bool (paired-end reads only)
            Whether to output reads that align discordantly
        contain: bool (paired-end reads only)
            Whether to consider a mate that contains the other to be concordant
        dovetail: bool (paired-end reads only)
            Whether to consider dovetailed alignments to be concordant
        mixed: bool (paired-end reads only)
            Whether to output a mate that aligns if the other mate does not

        Returns
        -------
        RefsetAlignmentFilePath | OneRefAlignmentFilePath
            Path of the output SAM file
        """
        cmd = [BOWTIE2_CMD]
        # Resources
        cmd.extend(["-p", self.max_cpus])
        # Alignment
        if local:
            cmd.append("--local")
        if gap_bar:
            cmd.extend(["--gbar", gap_bar])
        if padding:
            cmd.extend(["--dpad", padding])
        if seed_size:
            cmd.extend(["-L", seed_size])
        if seed_interval:
            cmd.extend(["-i", seed_interval])
        if extensions:
            cmd.extend(["-D", extensions])
        if reseed:
            cmd.extend(["-R", reseed])
        # Scoring
        cmd.append(self.fastq.phred_arg)
        cmd.append("--ignore-quals")
        cmd.extend(["--ma", MATCH_BONUS])
        cmd.extend(["--mp", MISMATCH_PENALTY])
        cmd.extend(["--np", N_PENALTY])
        cmd.extend(["--rfg", REF_GAP_PENALTY])
        cmd.extend(["--rdg", READ_GAP_PENALTY])
        # Filtering
        if score_min:
            cmd.extend(["--score-min", score_min])
        if frag_len_min:
            cmd.extend(["-I", frag_len_min])
        if frag_len_max:
            cmd.extend(["-X", frag_len_max])
        if not unaligned:
            cmd.append("--no-unal")
        # Mate pair orientation
        if orientation in set(MateOrientation):
            cmd.append(f"--{orientation}")
        if not discordant:
            cmd.append("--no-discordant")
        if not contain:
            cmd.append("--no-contain")
        if dovetail:
            cmd.append("--dovetail")
        if not mixed:
            cmd.append("--no-mixed")
        # Formatting
        cmd.append("--xeq")
        # Inputs and outputs
        cmd.extend(["-S", self.output.path])
        cmd.extend(["-x", self.fasta_prefix])
        cmd.extend(self.fastq.bowtie2_inputs)
        # Run alignment
        run_cmd(cmd)
        return self.output

    def run(self, **kwargs):
        self.setup()
        self._bowtie2_build()
        return self._bowtie2(**kwargs)

    def clean(self):
        self.output.path.unlink()


class XamBase(ReadsFileBase):
    def __init__(self,
                 top_dir: path.TopDirPath,
                 max_cpus: int,
                 xam: path.RefsetAlignmentInFilePath |
                      path.OneRefAlignmentInFilePath):
        super().__init__(top_dir, max_cpus)
        self.xam = xam

    @property
    def sample(self):
        return self.xam.sample

    @property
    def demult(self):
        if isinstance(self.xam, path.RefsetAlignmentInFilePath):
            return False
        if isinstance(self.xam, path.OneRefAlignmentInFilePath):
            return True
        raise TypeError(self.xam)

    @cached_property
    def output(self):
        if self.partition == path.Partition.TEMP:
            trans = path.AlignmentInToAlignmentTemp
        elif self.partition == path.Partition.OUTPUT:
            trans = path.AlignmentInToAlignmentOut
        else:
            raise ValueError(self.partition)
        return trans.trans_inst(self.xam,
                                **self.fields,
                                ext=self.ext,
                                preserve_type=True)

    @staticmethod
    def _build_index(xam: path.TopDirPath):
        cmd = [SAMTOOLS_CMD, "index", xam.path]
        run_cmd(cmd)
        xam_index = xam.replace(ext=path.BAI_EXT)
        if not xam_index.path.is_file():
            raise FileNotFoundError(xam_index.path)
        return xam_index

    def index_input(self):
        return self._build_index(self.xam)

    def index_output(self):
        return self._build_index(self.output)

    def clean(self):
        self.output.path.unlink(missing_ok=True)


class SamRemoveEqualMappers(XamBase):
    step = path.TempStep.ALIGN_REMEQ
    ext = path.SAM_EXT

    pattern_a = re.compile(SAM_ALIGN_SCORE + rb"(\d+)")
    pattern_x = re.compile(SAM_EXTRA_SCORE + rb"(\d+)")

    _MIN_SAM_FIELDS = 11
    _MAX_SAM_FLAG = 4095  # 2^12 - 1

    @staticmethod
    def _get_score(line: bytes, ptn: re.Pattern[bytes]):
        return (float(match.groups()[0])
                if (match := ptn.search(line)) else None)

    @classmethod
    def _is_best_alignment(cls, line: bytes):
        return ((score_x := cls._get_score(line, cls.pattern_x)) is None
                or score_x < cls._get_score(line, cls.pattern_a))

    @classmethod
    def _read_is_paired(cls, line: bytes):
        info = line.split()
        if len(info) < cls._MIN_SAM_FIELDS:
            raise ValueError(f"Invalid SAM line:\n{line.decode()}")
        flag = int(info[1])
        if 0 <= flag <= cls._MAX_SAM_FLAG:
            return bool(flag % 2)
        raise ValueError(f"Invalid SAM flag: {flag}")

    def _iter_paired(self, sam: BinaryIO, line: bytes):
        for line2 in sam:
            if self._is_best_alignment(line) or self._is_best_alignment(line2):
                yield b"".join((line, line2))
            line = sam.readline()

    def _iter_single(self, sam: BinaryIO, line: bytes):
        while line:
            if self._is_best_alignment(line):
                yield line
            line = sam.readline()

    def _remove_equal_mappers(self, buffer_length=BUFFER_LENGTH):
        with (open(self.xam.path, "rb") as sami,
              open(self.output.path, "wb") as samo):
            # Copy the header from the input to the output SAM file.
            while (line := sami.readline()).startswith(SAM_HEADER):
                samo.write(line)
            if line:
                if self._read_is_paired(line):
                    lines = self._iter_paired(sami, line)
                else:
                    lines = self._iter_single(sami, line)
                while text := b"".join(itertools.islice(lines, buffer_length)):
                    samo.write(text)
        return self.output

    def run(self):
        logging.info("\nRemoving Reads Mapping Equally to Multiple Locations"
                     f" in {self.xam.path}\n")
        self.setup()
        return self._remove_equal_mappers()


class XamSorter(XamBase):
    def _sort(self, name: bool):
        cmd = [SAMTOOLS_CMD, "sort"]
        if name:
            cmd.append("-n")
        cmd.extend(["-o", self.output.path, self.xam.path])
        run_cmd(cmd)
        return self.output

    def run(self, name: bool = False):
        logging.info(f"\nSorting {self.xam.path} by Reference and Coordinate\n")
        self.setup()
        return self._sort(name)


class BamAlignSorter(XamSorter):
    step = path.TempStep.ALIGN_SORT
    ext = path.BAM_EXT


class SamVectorSorter(XamSorter):
    module = path.Module.VECTOR
    step = path.TempStep.VECTOR_SORT
    ext = path.SAM_EXT


class BamSplitter(XamBase):
    partition = path.Partition.OUTPUT
    step = path.TempStep.ALIGN_SPLIT
    ext = path.BAM_EXT

    def __init__(self,
                 top_dir: path.TopDirPath,
                 max_cpus: int,
                 xam: path.RefsetAlignmentInFilePath |
                      path.OneRefAlignmentInFilePath,
                 fasta: path.RefsetSeqInFilePath |
                        path.OneRefSeqTempFilePath):
        super().__init__(top_dir, max_cpus, xam)
        if isinstance(fasta, path.RefsetSeqInFilePath):
            if self.demult:
                raise TypeError("Got multi-FASTA but demultiplexed BAM")
        elif isinstance(fasta, path.OneRefSeqTempFilePath):
            if not self.demult:
                raise TypeError("Got single-FASTA but multiplexed BAM")
        else:
            raise TypeError(fasta)
        self.fasta = fasta

    @cached_property
    def refs(self):
        return tuple(ref for ref, _ in FastaParser(self.fasta.path).parse())

    def _get_cmd(self, output: pathlib.Path, ref: str = ""):
        cmd = [SAMTOOLS_CMD, "view"]
        if self.ext == path.BAM_EXT:
            cmd.append("-b")
        cmd.extend(("-o", output, self.xam.path))
        if ref:
            cmd.append(ref)
        return cmd

    def _extract_one_ref(self, ref: str):
        output = path.OneRefAlignmentOutFilePath(**self.fields, ref=ref,
                                                 ext=self.ext)
        cmd = self._get_cmd(output.path, ref=ref)
        run_cmd(cmd)
        return output

    def _split_xam(self):
        self.index_input()
        return tuple(map(self._extract_one_ref, self.refs))

    def _move_xam(self):
        if self.ext == self.xam.ext:
            os.rename(self.xam.path, self.output.path)
        else:
            cmd = self._get_cmd(self.output.path)
            run_cmd(cmd)
        return self.output,

    def _build_outputs(self):
        if self.demult:
            return self._move_xam()
        else:
            return self._split_xam()

    def _build_indexes(self, bams: tuple[path.TopDirPath]):
        return tuple(map(self._build_index, bams))

    def run(self):
        logging.info(f"\nSplitting {self.xam.path} by reference\n")
        self.setup()
        bams = self._build_outputs()
        self._build_indexes(bams)
        return bams

    def clean(self):
        return


class BamVectorSelector(XamBase):
    module = path.Module.VECTOR
    step = path.TempStep.VECTOR_SELECT
    ext = path.BAM_EXT

    def __init__(self,
                 top_dir: path.TopDirPath,
                 max_cpus: int,
                 xam: path.OneRefAlignmentInFilePath,
                 ref: str,
                 end5: int,
                 end3: int):
        super().__init__(top_dir, max_cpus, xam)
        self.ref = ref
        self.end5 = end5
        self.end3 = end3

    @staticmethod
    def ref_coords(ref: str, end5: int, end3: int):
        return f"{ref}:{end5}-{end3}"

    @property
    def fields(self):
        return {**super().fields, "ref": self.ref}

    @cached_property
    def output_dir(self):
        if self.partition == path.Partition.TEMP:
            return path.RefTempDirPath(**self.fields)
        if self.partition == path.Partition.OUTPUT:
            return path.RefOutDirPath(**self.fields)
        raise ValueError(self.partition)

    @cached_property
    def output(self):
        if self.partition == path.Partition.TEMP:
            trans = path.AlignmentInToRegionAlignmentTemp
        elif self.partition == path.Partition.OUTPUT:
            trans = path.AlignmentInToRegionAlignmentOut
        else:
            raise ValueError(self.partition)
        return trans.trans_inst(self.xam,
                                **self.fields,
                                end5=self.end5,
                                end3=self.end3,
                                ext=self.ext,
                                preserve_type=True)

    def _select(self):
        cmd = [SAMTOOLS_CMD, "view", "-h", "-o", self.output.path,
               self.xam.path, self.ref_coords(self.ref, self.end5, self.end3)]
        run_cmd(cmd)
        return self.output

    def run(self):
        self.setup()
        return self._select()


'''
primer1 = "CAGCACTCAGAGCTAATACGACTCACTATA"
primer1rc = "TATAGTGAGTCGTATTAGCTCTGAGTGCTG"
primer2 = "TGAAGAGCTGGAACGCTTCACTGA"
primer2rc = "TCAGTGAAGCGTTCCAGCTCTTCA"
adapters5 = (primer1, primer2rc)
adapters3 = (primer2, primer1rc)
'''



from __future__ import annotations
from functools import cached_property, wraps
from io import BufferedReader
from typing import Callable, Optional

from dreem.util.reads import BamVectorSelector, SamVectorSorter
from dreem.util.path import TopDirPath, OneRefAlignmentInFilePath, OneRefAlignmentTempFilePath
from dreem.vector.vector import *


def _reset_seek(func: Callable):
    @wraps(func)
    def wrapper(samview: SamViewer, *args, **kwargs):
        prev_pos = samview.sam_file.tell()
        result = func(samview, *args, **kwargs)
        samview.sam_file.seek(prev_pos)
        return result
    return wrapper


def _range_of_records(func: Callable):
    @wraps(func)
    @_reset_seek
    def wrapper(samview: SamViewer, start: int, stop: int):
        samview.sam_file.seek(start)
        records = iter(func(samview))
        if stop is None:
            return records
        else:
            while True:
                if samview.sam_file.tell() < stop:
                    yield next(records)
                else:
                    assert samview.sam_file.tell() == stop
                    break
    return wrapper


class SamViewer(object):
    def __init__(self,
                 top_dir: TopDirPath,
                 max_cpus: int,
                 xam_path: OneRefAlignmentInFilePath,
                 ref_name: str,
                 end5: int,
                 end3: int,
                 spanning: bool,
                 min_qual: int,
                 owner: bool = True):
        self.top_dir = top_dir
        self.max_cpus = max_cpus
        self.xam_path = xam_path
        self.ref_name = ref_name
        self.end5 = end5
        self.end3 = end3
        self.spanning = spanning
        self.min_qual = min_qual
        self.owner = owner
        self._sam_path: OneRefAlignmentTempFilePath | None = None
        self._sam_file: BufferedReader | None = None
    
    def __enter__(self):
        # Convert the BAM file to a temporary SAM file
        if self.owner:
            if self.spanning:
                selector = None
                xam_selected = self.xam_path
            else:
                selector = BamVectorSelector(self.top_dir,
                                             self.max_cpus,
                                             self.xam_path,
                                             self.ref_name,
                                             self.end5,
                                             self.end3)
                xam_selected = selector.run()
            sorter = SamVectorSorter(self.top_dir, self.max_cpus, xam_selected)
            self._sam_path = sorter.run(name=True)
            if selector:
                selector.clean()
        else:
            self._sam_path = self.xam_path
        self._sam_file = open(self.sam_path.path, "rb")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._sam_file.close()
        self._sam_file = None
        if self.owner:
            self.sam_path.path.unlink()
            self._sam_path = None
    
    @property
    def sam_path(self):
        if not self._sam_path:
            raise RuntimeError(f"{self} has no open SAM path.")
        return self._sam_path

    @property
    def sam_file(self):
        if not self._sam_file:
            raise RuntimeError(f"{self} has no open SAM file.")
        return self._sam_file

    def _seek_beginning(self):
        self.sam_file.seek(0)
    
    @cached_property
    @_reset_seek
    def _rec1_pos(self):
        self._seek_beginning()
        while (line := self.sam_file.readline()).startswith(SAM_HEADER):
            pass
        return self.sam_file.tell() - len(line)

    def _seek_rec1(self):
        self.sam_file.seek(self._rec1_pos)
        
    @cached_property
    @_reset_seek
    def paired(self):
        self._seek_rec1()
        first_line = self.sam_file.readline()
        return (SamRead(first_line, self.min_qual).flag.paired if first_line
                else False)
    
    @_range_of_records
    @staticmethod
    def _get_records_single(samview: SamViewer):
        while read := SamRead(samview.sam_file.readline(), samview.min_qual):
            if read.flag.paired:
                raise ValueError("Found paired-end read in single-end SAM file.")
            yield SamRecord(read)

    @_range_of_records
    @staticmethod
    def _get_records_paired_flexible(samview: SamViewer):
        prev_read: Optional[SamRead] = None
        while line := samview.sam_file.readline():
            read = SamRead(line, samview.min_qual)
            assert read.flag.paired
            if prev_read:
                # The previous read has not yet been yielded
                if prev_read.qname == read.qname:
                    # The current read is the mate of the previous read
                    if prev_read.flag.end5:
                        yield SamRecord(prev_read, read)
                    else:
                        yield SamRecord(read, prev_read)
                    prev_read = None
                else:
                    # The previous read is paired, but its mate is not in
                    # the SAM file. This situation can happen when only the
                    # reads mapping to a specific region are exported to a
                    # SAM file and only one mate overlaps that region.
                    yield SamRecord(prev_read)
                    # Save the current read so that if its mate is the next
                    # read, it will be returned as a pair.
                    prev_read = read
            else:
                # Save the current read so that if its mate is the next
                # read, it will be returned as a pair.
                prev_read = read
        if prev_read:
            # In case the last record
            yield SamRecord(prev_read)
    
    @_range_of_records
    @staticmethod
    def _get_records_paired_strict(samview: SamViewer):
        while line := samview.sam_file.readline():
            yield SamRecord(SamRead(line,
                                    samview.min_qual),
                            SamRead(samview.sam_file.readline(),
                                    samview.min_qual))
    
    def get_records(self, start: int, stop: int):
        if self.paired:
            if self.spanning:
                record_generator = self.__class__._get_records_paired_strict
            else:
                record_generator = self.__class__._get_records_paired_flexible
        else:
            record_generator = self.__class__._get_records_single
        return record_generator(self, start, stop)
    
    @_reset_seek
    def get_batch_indexes(self, batch_size: int):
        if batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")
        n_skip = (self.paired + 1) * (batch_size - 1)
        self._seek_rec1()
        while True:
            yield self.sam_file.tell()
            try:
                line = next(self.sam_file)
            except StopIteration:
                break
            else:
                for _, line in zip(range(n_skip), self.sam_file):
                    pass
                if self.paired:
                    # Compare the current and next query names.
                    try:
                        qname = line.split()[0]
                    except IndexError:
                        pass
                    else:
                        line_next = self.sam_file.readline()
                        try:
                            qname_next = line_next.split()[0]
                        except IndexError:
                            pass
                        else:
                            if qname != qname_next:
                                # If the current and next query names differ
                                # (the lines do not come from two paired mates),
                                # then backtrack to the beginning of line_next.
                                self.sam_file.seek(-len(line_next), 1)



import click
from dreem.align.main import run
from dreem.util.cli import *

@click.command()
@arg_fasta
@opt_top_dir
@opt_fastqs
@opt_fastqi
@opt_fastq1
@opt_fastq2
@opt_fastqs_dir
@opt_fastqi_dir
@opt_fastq12_dir
@opt_phred_enc
@opt_parallel
@opt_cpus
def cli(*args, **kwargs):
    run(*args, **kwargs)


if __name__ == '__main__':
    cli()



from collections import Counter
import pandas as pd
import numpy as np
from scipy import signal
import datetime
from dreem.util.files_sanity import check_library
from dreem.util import path
from dreem.util.reads import FastqUnit

def demultiplex(fq_unit: FastqUnit,
                fasta: path.RefsetSeqInFilePath,
                out_dir: path.ModuleDirPath,
                library: str,
                max_barcode_mismatches: int):
    """Demultiplex a pair of FASTQ files.

    Publishes to `output_folder` a pair of FASTQ files for each construct, named {construct}_R1.fastq and {construct}_R2.fastq.

    Parameters
    ----------
    fq_unit: FastqUnit
        One single-end or interleaved paired-end FASTQ file, or paired-end
        reads with mates 1 and 2 in two separate FASTQ files.
    fasta: RefsetSeqInFilePath
        FASTA file containing the reference sequences.
    library: pd.DataFrame
        Columns are (non-exclusively): ['construct', 'barcode_start', 'barcode']
    out_dir: str
        Where to output the results.
    max_barcode_mismatches: int
        Maximum number of mutations allowed on the barcode.
        
    Returns
    -------
    dict[str, FastqUnit]
        Dictionary mapping construct names to the demultiplexed FASTQ files.
    """

    out_dir.path.mkdir(parents=True, exist_ok=True)

    # Remove the report file if it exists
    report_path = out_dir.path.joinpath('report.txt')
    if report_path.is_file():
        report_path.unlink()
    
    library = check_library(pd.read_csv(library), str(fasta.path))
    constructs = library['construct'].unique()
    barcodes = library['barcode'].unique()

    assert len(constructs) == len(barcodes)
    for construct, barcode in zip(constructs, barcodes):
        # check if the barcode and the construct are on the same row
        assert library.loc[library['construct']==construct, 'barcode'].values[0] == barcode

    lost_reads = fq_unit.trans(path.ReadsInToReadsOut, ref="lost_reads",
                               **out_dir.dict())

    construct_fastqs: dict[str, FastqUnit] = dict()

    # copy the reads from the fastq files that contain the barcode in a fastq file named after the construct in the output folder
    for second, fq in enumerate(fq_unit.inputs):
        
        # infos for the report
        perfect_matches_count = 0
        off_matches_count = {k:0 for k in range(1, max_barcode_mismatches+1)}
        lost_reads_count = 0
        count_per_construct = {construct:0 for construct in constructs}
        barcode_shifts = []
                        
        with open(fq.path, 'r') as f:
            while True:
                header, sequence, quality = read_fastq_line(f)
                if not header:
                    break

                for construct, barcode in zip(constructs, barcodes):
                    
                    minimal_corr_score = worst_matching_score(barcode, max_barcode_mismatches)
                    
                    if second:
                        barcode = reverse_complement(barcode)
                    
                    corr = compute_correlation(embed_sequence_as_binary(barcode), embed_sequence_as_binary(sequence))
                    is_match = barcode_in_read(corr, minimal_corr_score)
                    if is_match:
                        if max(corr) == 1:
                            perfect_matches_count += 1
                        else:
                            for k in range(1, max_barcode_mismatches+1):
                                if max(corr) >= worst_matching_score(barcode, k):
                                    off_matches_count[k] += 1
                                    break
                            
                        barcode_start = library.loc[library['construct']==construct, 'barcode_start'].values[0]
                        if second:
                            barcode_shifts.append(len(sequence) - np.argmax(corr) - barcode_start - len(barcode))
                        else:
                            barcode_shifts.append(np.argmax(corr) - barcode_start)
                            
                        count_per_construct[construct] += 1

                        # TODO: Open each construct's file a minimum number of times (not once per line)
                        try:
                            with open(construct_fastqs[construct].paths[second], "a") as g:
                                write_fastq_line(g, header, sequence, quality)
                        except KeyError:
                            construct_fq_unit = fq_unit.trans(path.ReadsInToReadsOut,
                                                              ref=construct,
                                                              **out_dir.dict())
                            construct_fastqs[construct] = construct_fq_unit
                            with open(construct_fq_unit.paths[second], "w") as g:
                                write_fastq_line(g, header, sequence, quality)
                        break
                else:
                    lost_reads_count += 1
                    with open(lost_reads.paths[second], 'a') as g:
                        write_fastq_line(g, header, sequence, quality)
        write_report(fq.path, report_path, perfect_matches_count, off_matches_count, lost_reads_count, barcode_shifts, count_per_construct)
    return construct_fastqs

def write_report(fastq, report_path, perfect_matches_count, off_matches_count, lost_reads_count, barcode_shifts, count_per_construct):
    """Write a report of the demultiplexing process for the given fastq file."""
    with open(report_path, 'a') as f:
        f.write("Time: " + str(datetime.datetime.now()) + "\n")
        f.write('\n'+'='*len('Demultiplexing report for ' + fastq) + '\n')
        f.write('Demultiplexing report for ' + fastq + '\n')
        f.write('='*len('Demultiplexing report for ' + fastq) + '\n')
        f.write('Count of perfect matches: ' + str(perfect_matches_count) + '\n')
        for k in off_matches_count:
            f.write('Count of ' + str(k) + '-off matches: ' + str(off_matches_count[k]) + '\n')
        f.write('Count of lost reads: ' + str(lost_reads_count) + '\n')
        f.write('Count of reads per barcode position: ' + str(bin_positions(barcode_shifts)) + '\n')
        f.write('\nCount of reads per construct: ' + '\n' + '-'*len('Count of reads per construct:') + '\n')
        for construct in count_per_construct:
            f.write(construct + ': ' + str(count_per_construct[construct]) + '\n')
        f.write('='*len('Demultiplexing report for ' + fastq) + '\n')
        
def worst_matching_score(barcode, max_muts=1):
    return 1. - float(max_muts)/len(barcode) -1E-9
                                    
def bin_positions(positions):
    """Turns a list of positions into a dictionary of bins."""
    bins = {}
    for pos in positions:
        if pos not in bins:
            bins[pos] = 0
        bins[pos] += 1
    bins = {k: bins[k] for k in sorted(bins)}
    return bins

def read_fastq_line(f):
    """Read the line of a fastq file and return a tuple (header, sequence, quality)"""
    header = f.readline().strip()
    sequence = f.readline().strip()
    f.readline()
    quality = f.readline().strip()
    return header, sequence, quality

def write_fastq_line(f, header, sequence, quality):
    """Write a fastq file line."""
    f.write(header + '\n')
    f.write(sequence + '\n')
    f.write('+\n')
    f.write(quality + '\n')

def embed_sequence_as_binary(sequence):
    """Each sequence is represented as 4 binary vector of length len(sequence), one per base A C T G."""
    return np.array([np.array([base == 'A', base == 'C', base == 'T', base == 'G']) for base in sequence], dtype=np.int8).T

def compute_correlation(barcode, read):
    """ Use the correlation between the barcode and the read to determine if the barcode is in the read."""
    return signal.correlate(read,barcode, mode="valid", method="auto").squeeze()/barcode.shape[1]

def barcode_in_read(corr, min_corr_score):
    """Return True if the correlation score is above the threshold, False otherwise."""
    return max(corr) > min_corr_score

def hamming_distance(s1, s2):
    assert len(s1) == len(s2)
    return sum(ch1 != ch2 for ch1, ch2 in zip(s1, s2))

def reverse_complement(seq):
    return seq[::-1].translate(str.maketrans('ATCG', 'TAGC'))

def next_base(base):
    return {'A':'T','T':'C','C':'G','G':'A',0:1}[base]

def run(top_dir: str, fasta: str, phred_enc: int,
        fastqs: tuple[str], fastqi: tuple[str],
        fastq1: tuple[str], fastq2: tuple[str],
        library: str, max_barcode_mismatches: int):
    """Run the demultiplexing pipeline.

    Demultiplexes the reads and outputs one fastq file per construct in the directory `output_path`, using `temp_path` as a temp directory.

    Parameters from args:
    -----------------------
    fastq1: str
        Path to the FASTQ file or list of paths to the FASTQ files, forward primer.
    fastq2: str
        Path to the FASTQ file or list of paths to the FASTQ files, reverse primer.
    fasta: str
        Path to the FASTA file.
    interleaved: bool
        If True, the FASTQ files are interleaved.
    library: str
        Path to the library file. Columns are (non-excusively): ['construct', 'barcode_start', 'barcode']
    out_dir: str
        Name of the output directory.
    max_barcode_mismatches: int
        Maximum number of mutations allowed on the barcode.
    verbose: bool
        Print progress to stdout (default: no).
        
    Returns
    -------
    dict[str, dict[str, FastqUnit]]
        Dictionary mapping sample names to

    """

    fasta_path = path.RefsetSeqInFilePath.parse_path(fasta)
    # Make the folders
    out_dir = path.ModuleDirPath(top=top_dir,
                                 partition=path.Partition.OUTPUT,
                                 module=path.Module.DEMULT)

    # Demultiplex.
    demultiplexed: dict[str, dict[str, FastqUnit]] = dict()
    fq_units = FastqUnit.from_strs(fastqs=fastqs, fastqi=fastqi,
                                   fastq1=fastq1, fastq2=fastq2,
                                   phred_enc=phred_enc, demult=True)
    # Ensure that no sample names are duplicated.
    if dups := [sample for sample, count
                in Counter(fq.sample for fq in fq_units).items()
                if count > 1]:
        raise ValueError(f"Got duplicate sample names: {', '.join(dups)}")

    # TODO: Parallelize with multiprocessing.Pool.starmap
    for fq_unit in fq_units:
        demultiplexed[fq_unit.sample] = demultiplex(
            fq_unit=fq_unit,
            fasta=fasta_path,
            out_dir=out_dir,
            library=library,
            max_barcode_mismatches=max_barcode_mismatches)
    return demultiplexed



from dreem.align import align
from dreem.util.reads import FastqUnit


def run(parallel: str,
        max_cpus: int,
        top_dir: str,
        fasta: str,
        phred_enc: int,
        fastqs: tuple[str] = (),
        fastqi: tuple[str] = (),
        fastq1: tuple[str] = (),
        fastq2: tuple[str] = (),
        fastqs_dir: tuple[str] = (),
        fastqi_dir: tuple[str] = (),
        fastq12_dir: tuple[str] = (),
        **kwargs):
    """
    Run the alignment module.

    Align the reads to the set of reference sequences and output one BAM file
    for each sample aligned to each reference in the directory 'output'.
    Temporary intermediary files are written in the directory 'temp' and then
    deleted after they are no longer needed.

    {top_dir}/
        output/
            {sample_1}/
                {ref_1}.bam
                {ref_2}.bam
                ...
            {sample_2}/
            ...
        temp/
            {step_1}/
                {sample_1}/
                    {ref_1}.file
                    {ref_2}.file
                    ...
                {sample_2}/
                ...
            {step_2}/
            ...
        ...

    Parameters
    ----------
    top_dir: str
        Path to the top-level folder into which all final outputs and temporary
        files are written. Must already exist.
    fasta: str
        Path to the reference FASTA file.
    fastqs: tuple[str] †
        Paths to all FASTQ files of single-end reads.
    fastqi: tuple[str] †
        Paths to all FASTQ files of interleaved, paired-end reads.
    fastq1: tuple[str] †
        Paths to all FASTQ files of mate 1 paired-end reads. If given, fastq2
        must also be given, and the sample names (in order) must match.
    fastq2: tuple[str] †
        Paths to all FASTQ files of mate 2 paired-end reads. If given, fastq1
        must also be given, and the sample names (in order) must match.
    fastqs_dir: tuple[str] ‡
        Paths to all directories containing demultiplexed FASTQ files of
        single-end reads.
    fastqi_dir: tuple[str] ‡
        Paths to all directories containing demultiplexed FASTQ files of
        interleaved paired-end reads.
    fastq12_dir: tuple[str] ‡
        Paths to all directories containing demultiplexed FASTQ files of
        mate 1 and mate 2 paired-end reads.
    phred_enc: int
        The ASCII encoding offset of the Phred scores. For example, if
        ```phred_enc = 33```, then a Phred score of 30 will be encoded as the
        ASCII character corresponding to 30 + 33 = 63, which is '?'; and in a
        FASTQ or SAM file, the character 'F' (ASCII value 70) denotes a Phred
        score of 70 - 33 = 37. This encoding of 33 is used by most modern
        Illumina sequencers and is the default.
    **kwargs
        Additional keyword arguments to pass to the alignment function.

    † The file name (minus the extension) will be used as the sample name.
    ‡ The file name (minus the extension) will be used as the reference name,
      and the name of the parent directory will be used as the sample name.
    """

    # Get all FASTQ units that need to be aligned.
    fq_units = list(FastqUnit.from_strs(phred_enc=phred_enc,
                                        fastqs=fastqs,
                                        fastqi=fastqi,
                                        fastq1=fastq1,
                                        fastq2=fastq2,
                                        fastqs_dir=fastqs_dir,
                                        fastqi_dir=fastqi_dir,
                                        fastq12_dir=fastq12_dir))

    # Run the alignment pipeline on each FASTQ.
    return align.run_steps_fqs(top_dir, fasta, fq_units, parallel, max_cpus,
                               **kwargs)



import os
from typing import List, Tuple

import pandas as pd

from dreem.util.util import DNA
from dreem.util.path import BAM_EXT
from dreem.vector.mprofile import VectorWriterSpawner
from dreem.util.files_sanity import check_library


def add_coords_from_library(library_path: str,
                            coords: List[Tuple[str, int, int]],
                            fasta: str,
                            out_dir: str):
    library = check_library(pd.read_csv(library_path), fasta, out_dir)
    for construct, end5, end3 in zip(library["construct"],
                                     library["section_start"],
                                     library["section_end"], strict=True):
        coord = (str(os.path.splitext(construct)[0]), int(end5), int(end3))
        if coord not in coords:
            coords.append(coord)


def encode_primers(primers: List[Tuple[str, str, str]]):
    return [(ref, DNA(fwd.encode()), DNA(rev.encode()))
            for ref, fwd, rev in primers]


def run(fasta: str, bam_dirs: list[str], phred_enc: int, min_phred: int,
        library: str, coords: list, primers: list, fill: bool, parallel: str,
        max_cpus: int, top_dir: str, rerun: bool):
    """
    Run the vectoring step.
    Generate a vector encoding mutations for each read
    (or read pair, if paired-end).

    FASTA (path): reference sequence(s) in FASTA format

    BAM_FILES (paths): list of one or more alignment files (space separated)
                       in BAM format
    """

    # read library
    if library:
        add_coords_from_library(library_path=library,
                                coords=coords,
                                fasta=fasta,
                                out_dir=top_dir)
                        
    # Compute mutation vectors for each BAM file
    bam_files = [os.path.join(bam_dir, bam_file)
                 for bam_dir in set(bam_dirs)
                 for bam_file in os.listdir(bam_dir)
                 if bam_file.endswith(BAM_EXT)]
    writers = VectorWriterSpawner(top_dir=top_dir,
                                  fasta=fasta,
                                  bam_files=bam_files,
                                  coords=coords,
                                  primers=encode_primers(primers),
                                  fill=fill,
                                  parallel=parallel,
                                  max_cpus=max_cpus,
                                  min_phred=min_phred,
                                  phred_enc=phred_enc,
                                  rerun=rerun)
    return writers.generate_profiles()
